#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#


import argparse
import importlib
import json
import logging
import os
from typing import Optional, List, Dict

from . import imp_exp
from .importer import SharedFolder, Team, Permission, PathDelimiter, replace_email_domain, BaseDownloadMembership
from .json.json import KeeperJsonImporter, KeeperJsonExporter
from ..commands.base import raise_parse_exception, suppress_exit, user_choice, Command
from ..params import KeeperParams


def register_commands(commands):
    commands['import'] = RecordImportCommand()
    commands['export'] = RecordExportCommand()
    commands['download-membership'] = DownloadMembershipCommand()
    commands['apply-membership'] = ApplyMembershipCommand()


def register_command_info(aliases, command_info):
    for p in [import_parser, export_parser, download_membership_parser, apply_membership_parser]:
        command_info[p.prog] = p.description


import_parser = argparse.ArgumentParser(prog='import', description='Import data from a local file into Keeper.')
import_parser.add_argument('--display-csv', '-dc', dest='display_csv', action='store_true',
                           help='display Keeper CSV import instructions')
import_parser.add_argument('--display-json', '-dj', dest='display_json', action='store_true',
                           help='display Keeper JSON import instructions')
import_parser.add_argument(
    '--format', choices=['json', 'csv', 'keepass', 'lastpass', 'myki', 'nordpass', 'manageengine', '1password',
                         'bitwarden', 'thycotic'],
    required=True, help='file format'
)
import_parser.add_argument('--folder', dest='folder', action='store',
                           help='import into a separate folder.')
import_parser.add_argument('-s', '--shared', dest='shared', action='store_true',
                           help='import folders as Keeper shared folders')
import_parser.add_argument('-p', '--permissions', dest='permissions', action='store',
                           help='default shared folder permissions: manage (U)sers, manage (R)ecords, can (E)dit, can (S)hare, or (A)ll, (N)one')
import_parser.add_argument('--update',  dest='update',  action='store_true',
                           help='Update records with common login, url or title')
import_parser.add_argument('--users',  dest='users',  action='store_true',
                           help='Update shared folder user permissions only')
import_parser.add_argument('--record-type', dest='record_type', action='store',
                           help='Import legacy records as record type')
import_parser.add_argument('--login-type', '-l', dest='login_type', action='store_true',
                           help='Import legacy records as login record type')
import_parser.add_argument('--old-domain', '-od', dest='old_domain', action='store',
                           help='old domain for changing user emails in permissions')
import_parser.add_argument('--new-domain', '-nd', dest='new_domain', action='store',
                           help='new domain for changing user emails in permissions')
import_parser.add_argument('--file-cache', dest='tmpdir', action='store',
                           help='Temp directory used to cache encrypted attachment imports')
import_parser.add_argument(
    'name', type=str, help='file name (json, csv, keepass, 1password), account name (lastpass), or URL (ManageEngine, Thycotic)'
)
import_parser.error = raise_parse_exception
import_parser.exit = suppress_exit


export_parser = argparse.ArgumentParser(prog='export', description='Export data from Keeper to a local file.')
export_parser.add_argument('--format', dest='format', choices=['json', 'csv', 'keepass'], required=True, help='file format')
export_parser.add_argument('--max-size', dest='max_size', help='Maximum file attachment file. Example: 100K, 50M, 2G. Default: 10M')
export_parser.add_argument('-kp', '--keepass-file-password', dest='keepass_file_password', action='store', help='Password for the exported Keepass file')
export_parser.add_argument('--force', dest='force', action='store_true', help='Suppress user interaction. Assume "yes"')
export_parser.add_argument('--folder', dest='folder', action='store', help='Export data from the specific folder only.')
export_parser.add_argument('name', type=str, nargs='?', help='file name or console output if omitted (except keepass)')
export_parser.error = raise_parse_exception
export_parser.exit = suppress_exit


download_membership_parser = argparse.ArgumentParser(prog='download-membership', description='Unload shared folder membership to JSON file.')
download_membership_parser.add_argument('--source', dest='source', choices=['keeper', 'lastpass', 'thycotic'], required=True, help='Shared folder membership source')
download_membership_parser.add_argument('--folder', dest='folder', action='store', help='import into a separate folder.')
download_membership_parser.add_argument('-p', '--permissions', dest='permissions', action='store', help='force shared folder permissions: manage (U)sers, manage (R)ecords')
download_membership_parser.add_argument('-r', '--restrictions', dest='restrictions', action='store', help='force shared folder restrictions: manage (U)sers, manage (R)ecords')
download_membership_parser.add_argument('--old-domain', '-od', dest='old_domain', action='store',  help='old domain for changing user emails in permissions')
download_membership_parser.add_argument('--new-domain', '-nd', dest='new_domain', action='store',  help='new domain for changing user emails in permissions')
download_membership_parser.add_argument('name', type=str, nargs='?', help='Output file name. "shared_folder_membership.json" if omitted.')
download_membership_parser.error = raise_parse_exception
download_membership_parser.exit = suppress_exit


apply_membership_parser = argparse.ArgumentParser(prog='apply-membership', description='Loads shared folder membership from JSON file into Keeper.')
apply_membership_parser.add_argument('name', type=str, nargs='?', help='Input file name. "shared_folder_membership.json" if omitted.')
apply_membership_parser.error = raise_parse_exception
apply_membership_parser.exit = suppress_exit

csv_instructions = '''CSV Import Instructions

File Format:
Folder,Title,Login,Password,Website Address,Notes,Shared Folder,Custom Fields

• To specify subfolders, use backslash "\\" between folder names
• To make a shared folder specify the name or path to it in the 7th field

Example 1: Create a regular folder at the root level with 2 custom fields
My Business Stuff,Twitter,marketing@company.com,123456,https://twitter.com,These are some notes,,API Key,5555,Date Created, 2018-04-02

Example 2: Create a shared subfolder inside another folder with edit and re-share permission
Personal,Twitter,craig@gmail.com,123456,https://twitter.com,,Social Media#edit#reshare

To load the sample data:
import --format=csv sample_data/import.csv
'''

json_instructions = '''JSON Import Instructions

Example JSON import file can be found in sample_data/import.json.txt.

The JSON file supports creating records, folders and shared folders.

Within shared folders, you can also automatically assign user or team permissions.

To load the sample file into your vault, run this command:
import --format=json sample_data/import.json.txt
'''


class ImporterCommand(Command):
    def execute_args(self, params, args, **kwargs):
        if args.find('--display-csv') >= 0 or args.find('-dc') >= 0:
            print(csv_instructions)
        elif args.find('--display-json') >= 0 or args.find('-dj') >= 0:
            print(json_instructions)
        else:
            Command.execute_args(self, params, args, **kwargs)


class RecordImportCommand(ImporterCommand):
    def get_parser(self):
        return import_parser

    def execute(self, params, **kwargs):
        if params.enforcements and 'booleans' in params.enforcements:
            restricted = next((x['value'] for x in params.enforcements['booleans'] if x['key'] == 'restrict_import'), False)
            if restricted:
                logging.warning('"import" is restricted by Keeper Administrator')
                return
        update_flag = kwargs['update'] if 'update' in kwargs else False
        import_format = kwargs['format'] if 'format' in kwargs else None
        import_name = kwargs['name'] if 'name' in kwargs else None
        shared = kwargs.get('shared') or False
        manage_users = False
        manage_records = False
        can_edit = False
        can_share = False
        if import_format and import_name:
            permissions = kwargs.get('permissions')
            if shared and not permissions:
                permissions = user_choice('Default shared folder permissions: manage (U)sers, manage (R)ecords, can (E)dit, can (S)hare, or (A)ll, (N)one', 'uresan', show_choice=False, multi_choice=True)
            if permissions:
                chars = set()
                chars.update([x for x in permissions.lower()])
                if 'a' in chars:
                    manage_users = True
                    manage_records = True
                    can_edit = True
                    can_share = True
                else:
                    if 'u' in chars:
                        manage_users = True
                    if 'r' in chars:
                        manage_records = True
                    if 'e' in chars:
                        can_edit = True
                    if 's' in chars:
                        can_share = True

            logging.info('Processing... please wait.')
            record_type = ''
            if kwargs.get('login_type'):
                record_type = 'login'
            rt = kwargs.get('record_type')
            if rt:
                if record_type and record_type != rt:
                    logging.warning('Options login-type and record-type are mutually exclusive.')
                    return
                record_type = rt
            if record_type:
                rti = None
                if params.record_type_cache:
                    for rts in params.record_type_cache.values():
                        try:
                            rto = json.loads(rts)
                            if rto.get('$id') == record_type:
                                rti = rto
                                break
                        except:
                            pass
                if rti is None:
                    logging.warning(f'Record type "{record_type}" not found.')
                    return

            imp_exp._import(params, import_format, import_name, shared=shared, import_into=kwargs.get('folder'),
                            manage_users=manage_users, manage_records=manage_records, users_only=kwargs.get('users') or False,
                            can_edit=can_edit, can_share=can_share, update_flag=update_flag, tmpdir=kwargs.get('tmpdir'),
                            old_domain=kwargs.get('old_domain'), new_domain=kwargs.get('new_domain'),
                            record_type=record_type)
        else:
            logging.error('Missing argument')


class RecordExportCommand(ImporterCommand):
    def get_parser(self):
        return export_parser

    def execute(self, params, **kwargs):

        if is_export_restricted(params):
            logging.warning('Permissions Required: `export` command is disabled. '
                            'Please contact your enterprise administrator.')
            return

        export_format = kwargs.pop('format', None)
        export_name = kwargs.pop('name', None)

        if format:
            msize = kwargs.pop('max_size', None)    # type: str
            if msize:
                multiplier = 1
                scale = msize[-1].upper()
                if scale == 'K':
                    multiplier = 1024
                elif scale == 'M':
                    multiplier = 1024 ** 2
                elif scale == 'G':
                    multiplier = 1024 ** 3

                if multiplier != 1:
                    msize = msize[:-1]
                try:
                    max_size = int(msize) * multiplier
                    kwargs['max_size'] = max_size
                except ValueError:
                    logging.error('Invalid maximum attachment file size parameter: %s', kwargs.get('max_size'))
                    return

            imp_exp.export(params, export_format, export_name, **kwargs)
        else:
            logging.error('Missing argument')


def is_export_restricted(params):
    is_export_restricted = False

    booleans = params.enforcements['booleans'] if params.enforcements and 'booleans' in params.enforcements else []

    if len(booleans) > 0:
        restrict_export_boolean = next((s for s in booleans if s['key'] == 'restrict_export'), None)

        if restrict_export_boolean:
            is_export_restricted = restrict_export_boolean['value']

    return is_export_restricted


def set_permission(perm, user, permit, restrict, perm_name):
    p = (user[perm_name] or permit[perm_name]) and not restrict[perm_name]
    setattr(perm, perm_name, p)


class DownloadMembershipCommand(Command):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return download_membership_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        source = kwargs.get('source') or 'keeper'
        file_name = kwargs.get('name') or 'shared_folder_membership.json'
        old_domain = kwargs.get('old_domain')
        new_domain = kwargs.get('new_domain')
        import_into = kwargs.get('folder')
        if import_into:
            import_into = import_into.replace(PathDelimiter, 2 * PathDelimiter)

        override_users = None    # Optional[bool]
        override_records = None  # Optional[bool]
        permissions = kwargs.get('permissions')
        if permissions:
            permissions = permissions.lower()
            if 'u' in permissions:
                override_users = True
            if 'r' in permissions:
                override_records = True

        restrictions = kwargs.get('restrictions')
        if restrictions:
            restrictions = restrictions.lower()
            if 'u' in restrictions:
                override_users = False
            if 'r' in restrictions:
                override_records = False

        added_folders = []  # type: List[SharedFolder]
        added_teams = []  # type: List[Team]

        try:
            full_name = f'keepercommander.importer.{"json" if source == "keeper" else source}'
            module = importlib.import_module(full_name)
            if hasattr(module, 'MembershipDownload'):
                plugin = module.MembershipDownload()   # type:   Optional[BaseDownloadMembership]
            else:
                logging.warning('Membership plugin is missing: %s', source)
                return
        except:
            logging.warning('Error loading membership plugin: %s', source)
            return

        for obj in plugin.download_membership(params):
            if isinstance(obj, SharedFolder):
                if import_into:
                    obj.path = f'{import_into}{PathDelimiter}{obj.path}'
                if isinstance(obj.permissions, list):
                    for p in obj.permissions:   # type: Permission
                        if old_domain and new_domain:
                            p.name = replace_email_domain(p.name, old_domain, new_domain)
                        if isinstance(override_users, bool):
                            p.manage_users = override_users
                        if isinstance(override_records, bool):
                            p.manage_records = override_records
                added_folders.append(obj)
            elif isinstance(obj, Team):
                if old_domain and new_domain and obj.members:
                    obj.members = [replace_email_domain(x, old_domain, new_domain) for x in obj.members]
                added_teams.append(obj)

        shared_folders = {}  # type: Dict[str, SharedFolder]
        teams = {}           # type: Dict[str, Team]

        json_importer = KeeperJsonImporter()
        if os.path.exists(file_name):
            try:
                for obj in json_importer.do_import(file_name, users_only=True):
                    if isinstance(obj, SharedFolder):
                        if obj.uid:
                            shared_folders[obj.uid] = obj
                    elif isinstance(obj, Team):
                        if obj.uid:
                            teams[obj.uid] = obj
            except:
                pass

        if added_folders or added_teams:
            for sf in added_folders:
                if sf.uid and sf.uid in shared_folders:
                    del shared_folders[sf.uid]
            for t in added_teams:
                if t.uid and t.uid in teams:
                    del teams[t.uid]

            memberships = []
            memberships.extend(shared_folders.values())
            memberships.extend(teams.values())
            memberships.extend(added_folders)
            memberships.extend(added_teams)
            json_exporter = KeeperJsonExporter()
            json_exporter.do_export(file_name, memberships)
            if len(added_folders) > 0:
                logging.info('%d shared folder memberships added.', len(added_folders))
            if len(added_teams) > 0:
                logging.info('%d team memberships added.', len(added_teams))
        else:
            logging.info('No folder memberships downloaded.')


class ApplyMembershipCommand(Command):
    def get_parser(self):
        return apply_membership_parser

    def execute(self, params, **kwargs):
        file_name = kwargs.get('name') or 'shared_folder_membership.json'
        if not os.path.exists(file_name):
            logging.warning('Shared folder membership file "%s" not found', file_name)
            return

        file_name = kwargs.get('name') or 'shared_folder_membership.json'

        shared_folders = []  # type: List[SharedFolder]
        teams = []     # type: List[Team]

        if os.path.exists(file_name):
            json_importer = KeeperJsonImporter()
            for obj in json_importer.do_import(file_name, users_only=True):
                if isinstance(obj, SharedFolder):
                    shared_folders.append(obj)
                if isinstance(obj, Team):
                    teams.append(obj)

        if len(shared_folders) > 0:
            imp_exp.import_user_permissions(params, shared_folders)

        if len(teams) > 0:
            imp_exp.import_teams(params, teams)



import json

from data_vault import VaultEnvironment
from keepercommander import api, crypto, utils
from keepercommander.params import KeeperParams

_TREE_KEY = api.generate_aes_key()
_ENTERPRISE_ID = 123

_VAULT_ENV = VaultEnvironment()

_USE_DATA_KEY = True

_TEAM_KEY = api.generate_aes_key()
_TEAM1_UID = api.generate_record_uid()
_TEAM2_UID = api.generate_record_uid()
_TEAM1_NAME = 'Team 1'
_TEAM2_NAME = 'Team 2'

_NODE1_ID = (_ENTERPRISE_ID << 32) + 2
_NODE2_ID = (_ENTERPRISE_ID << 32) + 100

_USER1_ID = (_ENTERPRISE_ID << 32) + 201
_USER2_ID = (_ENTERPRISE_ID << 32) + 202
_USER2_EMAIL = 'user2@keepercommander.com'

_ROLE1_ID = (_ENTERPRISE_ID << 32) + 301
_ROLE1_NAME = 'Role 1'

_LAST_ID = 1000


class EnterpriseEnvironment:
    def __init__(self):
        self.tree_key = _TREE_KEY
        self.team_key = _TEAM_KEY
        self.team1_uid = _TEAM1_UID
        self.team2_uid = _TEAM2_UID
        self.team1_name = _TEAM1_NAME
        self.team2_name = _TEAM2_NAME
        self.user1_id = _USER1_ID
        self.user2_id = _USER2_ID
        self.node1_id = _NODE1_ID
        self.node2_id = _NODE2_ID
        self.user2_email = _USER2_EMAIL
        self.role1_id = _ROLE1_ID
        self.role1_name = _ROLE1_NAME


def enterprise_allocate_ids(params, request):
    global _LAST_ID
    rs = {
        'result': 'success',
        'result_code': '',
        'message': ''
    }
    num = request['number_requested']
    rs['number_allocated'] = num
    rs['base_id'] = _LAST_ID
    _LAST_ID += num

    return rs


def get_enterprise_data(params):
    # type: (KeeperParams) -> dict

    encrypted_tree_key = crypto.encrypt_aes_v1(_TREE_KEY, params.data_key) \
        if _USE_DATA_KEY else crypto.encrypt_rsa(_TREE_KEY, _VAULT_ENV.public_key)
    tree_key_type = 1 if _USE_DATA_KEY else 2
    rs = {
        'result': 'success',
        'result_code': '',
        'message': '',
        'enterprise_name': 'Enterprise 1',
        'tree_key': utils.base64_url_encode(encrypted_tree_key),
        'key_type_id': tree_key_type
    }

    rs['nodes'] = [
        {
            'node_id':  _NODE1_ID,
            'encrypted_data': api.encrypt_aes(json.dumps({}).encode('utf-8'), _TREE_KEY)
        },
        {
            'node_id': _NODE2_ID,
            'parent_id': _NODE1_ID,
            'encrypted_data': api.encrypt_aes(json.dumps({'displayname': 'Sub node 1'}).encode('utf-8'), _TREE_KEY)
        }
    ]
    rs['users'] = [
        {
            'enterprise_user_id':  _USER1_ID,
            'node_id': _NODE1_ID,
            'username': params.user,
            'encrypted_data': api.encrypt_aes(json.dumps({'displayname': 'User 1'}).encode('utf-8'), _TREE_KEY),
            'status': 'active',
            'lock': 0
        },
        {
            'enterprise_user_id':  _USER2_ID,
            'node_id': _NODE2_ID,
            'username': _USER2_EMAIL,
            'encrypted_data': api.encrypt_aes(json.dumps({'displayname': 'User 2'}).encode('utf-8'), _TREE_KEY),
            'status': 'active',
            'lock': 1
        }

    ]
    rs['roles'] = [
        {
            'role_id': _ROLE1_ID,
            'node_id': _NODE1_ID,
            'encrypted_data': api.encrypt_aes(json.dumps({'displayname': _ROLE1_NAME}).encode('utf-8'), _TREE_KEY),
            'visible_below': True,
            'new_user_inherit': True
        }
    ]
    rs['managed_nodes'] = [
        {
            'role_id': _ROLE1_ID,
            'managed_node_id': _NODE1_ID,
            'cascade_node_management': True,
        }
    ]
    rs['role_users'] = [
        {
            'role_id': _ROLE1_ID,
            'enterprise_user_id':  _USER1_ID
        }
    ]
    rs['teams'] = [
        {
            'team_uid': _TEAM1_UID,
            'name': _TEAM1_NAME,
            'node_id': _NODE1_ID,
            'restrict_sharing': False,
            'restrict_edit': False,
            'restrict_view': False,
        },
        {
            'team_uid': _TEAM2_UID,
            'name': _TEAM2_NAME,
            'node_id': _NODE1_ID,
            'restrict_sharing': False,
            'restrict_edit': False,
            'restrict_view': False,
        }
    ]
    rs['team_users'] = [
        {
            'team_uid': _TEAM1_UID,
            'enterprise_user_id': _USER1_ID,
            'user_type': 1
        },
        {
            'team_uid': _TEAM2_UID,
            'enterprise_user_id': _USER2_ID,
            'user_type': 1
        }
    ]
    return rs



from collections import OrderedDict
import sys
from unittest import TestCase, mock

from keepercommander.commands import base
from keepercommander.cli import do_command
from data_vault import get_connected_params


class TestCommandLineInterface(TestCase):

    def test_command_import(self):
        commands = {}
        aliases = {}
        command_info = OrderedDict()

        #base.register_commands(commands, aliases, command_info)
        base.register_enterprise_commands(commands, aliases, command_info)

    def test_normalize_output_param(self):
        saved_platform = sys.platform
        try:
            # simulate windows platform
            sys.platform = 'win32'
            s = base.normalize_output_param(r'command --output=d:\1\2\aaa')
            self.assertEqual(s, r'command --output=d:/1/2/aaa')

            s = base.normalize_output_param(r'command --output d:\1\2\aaa')
            self.assertEqual(s, r'command --output d:/1/2/aaa')

            s = base.normalize_output_param(r'command d:\1\2\aaa')
            self.assertEqual(s, r'command d:\1\2\aaa')

            s = base.normalize_output_param(r'command --output d:/1/2\ 3/aaa')
            self.assertEqual(s, r'command --output d:/1/2\ 3/aaa')

            # simulate osx platform
            sys.platform = 'darwin'
            s = base.normalize_output_param(r'command --output=d:\1\2\aaa')
            self.assertEqual(s, r'command --output=d:\1\2\aaa')            
        finally:
            sys.platform = saved_platform

    def test_do_command_no_opts(self):
        params = get_connected_params()
        params.sync_data = False
        with mock.patch('keepercommander.commands.utils.ThisDeviceCommand.print_device_info') \
                as mock_print_dev:
            mock_print_dev.return_value = 'test device info'
            do_command(params, 'this-device')
            mock_print_dev.assert_called()



from setuptools import setup, find_packages
from os import path

import keepercommander

here = path.abspath(path.dirname(__file__))

# Get the long description from the README.md file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

install_requires = [
    'wheel',
    'asciitree',
    'bcrypt',
    'colorama',
    'cryptography>=38.0.3',
    'paramiko',
    'paramiko_expect',
    'prompt_toolkit',
    'protobuf>=3.18.0',
    'pycryptodomex>=3.7.2',
    'pykeepass',
    'pyperclip',
    'pysocks',
    'requests',
    'tabulate',
    'keeper-secrets-manager-core>=16.2.0'
]
adpasswd_requires = ['ldap3']
aws_requires = ['boto3']
test_requires = ['pytest', 'testfixtures']

setup(
    name='keepercommander',
    version=keepercommander.__version__,
    description='Keeper Commander for Python 3',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Craig Lurey',
    author_email='craig@keepersecurity.com',
    url='https://github.com/Keeper-Security/Commander',
    license='MIT',
    classifiers=["Development Status :: 4 - Beta",
                 "License :: OSI Approved :: MIT License",
                 "Operating System :: OS Independent",
                 "Programming Language :: Python :: 3.6",
                 "Topic :: Security"],
    keywords='security password',

    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3.6',
    entry_points={
        "console_scripts": [
            "keeper=keepercommander.__main__:main",
        ],
    },
    install_requires=install_requires,
    extras_require={
        'adpasswd': adpasswd_requires,
        'aws': aws_requires,
        'test': test_requires,
        'all': adpasswd_requires + aws_requires + test_requires
    }
)



import json
import os

import pytest

from keepercommander.params import KeeperParams
from keepercommander.error import KeeperApiError
from keepercommander import api, crypto
from keepercommander.commands.enterprise import EnterpriseTeamCommand, EnterpriseUserCommand, EnterpriseRoleCommand

from unittest import TestCase


@pytest.mark.integration
@pytest.mark.cross_enterprise
class TestCrossEnterpriseCommands(TestCase):
    params1 = None
    params2 = None

    @classmethod
    def setUpClass(cls):
        cls.params1 = KeeperParams()
        config_filename = os.path.join(os.path.dirname(__file__), 'cross-enterprise.json')
        with open(config_filename, 'r') as f:
            config = json.load(f)
            cls.params1.server = config['server']
            cls.params1.user = config['user']
            cls.params1.password = config['password']
            cls.params1.device_private_key = config['private_key']
            cls.params1.device_token = config['device_token']
            cls.params1.clone_code = config['clone_code']
            cls.params1.config = config
        api.login(cls.params1)

        cls.params2 = KeeperParams()
        cls.params2.server = cls.params1.server
        cls.params2.device_private_key = cls.params1.device_private_key
        cls.params2.device_token = cls.params1.device_token
        cls.params2.clone_code = cls.params1.clone_code
        cls.params2.user = config['enterprise2']['user']
        cls.params2.password = config['enterprise2']['password']
        api.login(cls.params2)

    def test_add_user_to_team(self):
        param1 = TestCrossEnterpriseCommands.params1     # type: KeeperParams
        param2 = TestCrossEnterpriseCommands.params2     # type: KeeperParams

        users = [x for x in param2.enterprise['users'] if x['username'] != param2.user]
        self.assertGreater(len(users), 0, 'cannot resolve user')
        user = users[0]
        ent2_user_id = user['enterprise_user_id']
        cmd = EnterpriseUserCommand()
        pk = cmd.get_public_key(param2, user['username'])

        #real team
        uids = []
        if 'teams' in param1.enterprise:
            uids = [x['team_uid'] for x in param1.enterprise['teams']]
        team_cmd = EnterpriseTeamCommand()
        if len(uids) == 0:
            team_cmd.execute(param1, add=True, team=['Team1'])
            uids = [x['team_uid'] for x in param1.enterprise['teams']]

        self.assertGreater(len(uids), 0, 'cannot resolve team')
        ent1_team_uid = uids[0]
        team_key = team_cmd.get_team_key(param1, ent1_team_uid)
        self.assertIsNotNone(team_key)
        rq = {
            "command": "team_enterprise_user_add",
            "team_uid": ent1_team_uid,
            "enterprise_user_id": ent2_user_id,
            "user_type": 0,
            "team_key": crypto.encrypt_rsa(team_key, pk)
        }
        failed = False
        try:
            api.communicate(param1, rq)
        except KeeperApiError as err:
            failed = True
            self.assertEqual(err.result_code, "bad_inputs_enterprise_user_id")
        self.assertTrue(failed)

        failed = False
        try:
            api.communicate(param2, rq)
        except KeeperApiError as err:
            failed = True
            self.assertEqual(err.result_code, "access_denied")
        self.assertTrue(failed)

        rq = {
            "command": "team_delete",
            "team_uid": ent1_team_uid
        }
        failed = False
        try:
            api.communicate(param2, rq)
        except KeeperApiError as err:
            failed = True
            self.assertEqual(err.result_code, "access_denied")
        self.assertTrue(failed)

    def test_add_user_to_role(self):
        param1 = TestCrossEnterpriseCommands.params1     # type: KeeperParams
        param2 = TestCrossEnterpriseCommands.params2     # type: KeeperParams

        users = [x for x in param2.enterprise['users']]
        self.assertGreater(len(users), 0, 'cannot resolve user')
        user = users[0]
        ent2_user_id = user['enterprise_user_id']

        ids = []
        if 'roles' in param1.enterprise:
            ids = [x['role_id'] for x in param1.enterprise['roles']]
        role_cmd = EnterpriseRoleCommand()
        if len(ids) == 0:
            role_cmd.execute(param1, add=True, role=['Role1'])
            ids = [x['role_id'] for x in param1.enterprise['roles']]

        self.assertGreater(len(ids), 0, 'cannot resolve role')
        ent1_role_id = ids[0]
        rq = {
            "command": "role_user_add",
            "role_id": ent1_role_id,
            "enterprise_user_id": ent2_user_id
        }
        failed = False
        try:
            api.communicate(param1, rq)
        except KeeperApiError as err:
            failed = True
            self.assertEqual(err.result_code, "bad_inputs_enterprise_user_id")
        self.assertTrue(failed)

        failed = False
        try:
            api.communicate(param2, rq)
        except KeeperApiError as err:
            failed = True
            self.assertEqual(err.result_code, "bad_inputs_role_id")
        self.assertTrue(failed)

        rq = {
            "command": "role_delete",
            "role_id": ent1_role_id
        }
        failed = False
        try:
            api.communicate(param2, rq)
        except KeeperApiError as err:
            failed = True
            self.assertEqual(err.result_code, "bad_inputs_role_id")
        self.assertTrue(failed)

    def test_lock_user(self):
        param1 = TestCrossEnterpriseCommands.params1     # type: KeeperParams
        param2 = TestCrossEnterpriseCommands.params2     # type: KeeperParams

        users = [x for x in param2.enterprise['users']]
        self.assertGreater(len(users), 0, 'cannot resolve user')
        user = users[0]
        ent2_user_id = user['enterprise_user_id']

        rq = {
            "command": "enterprise_user_lock",
            "enterprise_user_id": ent2_user_id,
            "lock": "locked",
            "delete_if_pending": True
        }
        failed = False
        try:
            api.communicate(param1, rq)
        except KeeperApiError as err:
            failed = True
            self.assertEqual(err.result_code, "bad_inputs_enterprise_user_id")
        self.assertTrue(failed)

        rq = {
            "command": "enterprise_user_delete",
            "enterprise_user_id": ent2_user_id
        }
        failed = False
        try:
            api.communicate(param1, rq)
        except KeeperApiError as err:
            failed = True
            self.assertEqual(err.result_code, "bad_inputs_enterprise_user_id")
        self.assertTrue(failed)

    def test_add_node(self):
        param1 = TestCrossEnterpriseCommands.params1     # type: KeeperParams
        param2 = TestCrossEnterpriseCommands.params2     # type: KeeperParams

        ent1_parent_id = [x['node_id'] for x in param1.enterprise['nodes']][0]

        cmd = EnterpriseRoleCommand()
        ent2_node_id = cmd.get_enterprise_id(param2)
        dt = { "displayname": "Node ID" }
        encrypted_data = api.encrypt_aes(json.dumps(dt).encode('utf-8'), param2.enterprise['unencrypted_tree_key'])
        rq = {
            "command": "node_add",
            "node_id": ent2_node_id,
            "parent_id": ent1_parent_id,
            "encrypted_data": encrypted_data
        }
        failed = False
        try:
            api.communicate(param2, rq)
        except KeeperApiError as err:
            failed = True
            self.assertEqual(err.result_code, "bad_inputs_parent_id")
        self.assertTrue(failed)

    def test_role_create(self):
        param1 = TestCrossEnterpriseCommands.params1 # type: KeeperParams
        param2 = TestCrossEnterpriseCommands.params2 # type: KeeperParams

        ent1_parent_id = [x['node_id'] for x in param1.enterprise['nodes']][0]

        cmd = EnterpriseRoleCommand()
        ent2_role_id = cmd.get_enterprise_id(param2)
        dt = { "displayname": "Role" }
        encrypted_data = api.encrypt_aes(json.dumps(dt).encode('utf-8'), param2.enterprise['unencrypted_tree_key'])
        rq = {
            "command": "role_add",
            "role_id": ent2_role_id,
            "node_id": ent1_parent_id,
            "encrypted_data": encrypted_data,
            "visible_below": True,
            "new_user_inherit": False
        }
        failed = False
        try:
            api.communicate(param2, rq)
        except KeeperApiError as err:
            failed = True
            self.assertEqual(err.result_code, "bad_inputs_node_id")
        self.assertTrue(failed)

    def test_team_get_key(self):
        param1 = TestCrossEnterpriseCommands.params1     # type: KeeperParams
        param2 = TestCrossEnterpriseCommands.params2     # type: KeeperParams

        uids = [x['team_uid'] for x in param1.enterprise['teams']]
        self.assertGreater(len(uids), 0, 'cannot resolve team')
        ent1_team_uid = uids[0]

        rq = {
            "command": "team_get_keys",
            "teams": [ent1_team_uid]
        }
        rs = api.communicate(param2, rq)
        self.assertEqual(rs['keys'][0]['result_code'], "doesnt_exist")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: AccountSummary.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import APIRequest_pb2 as APIRequest__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14\x41\x63\x63ountSummary.proto\x12\x0e\x41\x63\x63ountSummary\x1a\x10\x41PIRequest.proto\"/\n\x15\x41\x63\x63ountSummaryRequest\x12\x16\n\x0esummaryVersion\x18\x01 \x01(\x05\"\x9f\x04\n\x16\x41\x63\x63ountSummaryElements\x12\x11\n\tclientKey\x18\x01 \x01(\x0c\x12*\n\x08settings\x18\x02 \x01(\x0b\x32\x18.AccountSummary.Settings\x12*\n\x08keysInfo\x18\x03 \x01(\x0b\x32\x18.AccountSummary.KeysInfo\x12)\n\x08syncLogs\x18\x04 \x03(\x0b\x32\x17.AccountSummary.SyncLog\x12\x19\n\x11isEnterpriseAdmin\x18\x05 \x01(\x08\x12(\n\x07license\x18\x06 \x01(\x0b\x32\x17.AccountSummary.License\x12$\n\x05group\x18\x07 \x01(\x0b\x32\x15.AccountSummary.Group\x12\x32\n\x0c\x45nforcements\x18\x08 \x01(\x0b\x32\x1c.AccountSummary.Enforcements\x12(\n\x06Images\x18\t \x03(\x0b\x32\x18.AccountSummary.KeyValue\x12\x30\n\x0fpersonalLicense\x18\n \x01(\x0b\x32\x17.AccountSummary.License\x12\x1e\n\x16\x66ixSharedFolderRecords\x18\x0b \x01(\x08\x12\x11\n\tusernames\x18\x0c \x03(\t\x12+\n\x07\x64\x65vices\x18\r \x03(\x0b\x32\x1a.AccountSummary.DeviceInfo\x12\x14\n\x0cisShareAdmin\x18\x0e \x01(\x08\"\xb7\x02\n\nDeviceInfo\x12\x1c\n\x14\x65ncryptedDeviceToken\x18\x01 \x01(\x0c\x12\x12\n\ndeviceName\x18\x02 \x01(\t\x12\x32\n\x0c\x64\x65viceStatus\x18\x03 \x01(\x0e\x32\x1c.Authentication.DeviceStatus\x12\x17\n\x0f\x64\x65vicePublicKey\x18\x04 \x01(\x0c\x12 \n\x18\x65ncryptedDataKeyDoNotUse\x18\x05 \x01(\x0c\x12\x15\n\rclientVersion\x18\x06 \x01(\t\x12\x10\n\x08username\x18\x07 \x01(\t\x12\x11\n\tipAddress\x18\x08 \x01(\t\x12\x1a\n\x12\x61pproveRequestTime\x18\t \x01(\x03\x12\x1f\n\x17\x65ncryptedDataKeyPresent\x18\n \x01(\x08\x12\x0f\n\x07groupId\x18\x0b \x01(\x03\"\xc1\x01\n\x08KeysInfo\x12\x18\n\x10\x65ncryptionParams\x18\x01 \x01(\x0c\x12\x18\n\x10\x65ncryptedDataKey\x18\x02 \x01(\x0c\x12\x19\n\x11\x64\x61taKeyBackupDate\x18\x03 \x01(\x01\x12\x13\n\x0buserAuthUid\x18\x04 \x01(\x0c\x12\x1b\n\x13\x65ncryptedPrivateKey\x18\x05 \x01(\x0c\x12\x1e\n\x16\x65ncryptedEccPrivateKey\x18\x06 \x01(\x0c\x12\x14\n\x0c\x65\x63\x63PublicKey\x18\x07 \x01(\x0c\"\x81\x01\n\x07SyncLog\x12\x13\n\x0b\x63ountryName\x18\x01 \x01(\t\x12\x12\n\nsecondsAgo\x18\x02 \x01(\x03\x12\x12\n\ndeviceName\x18\x03 \x01(\t\x12\x13\n\x0b\x63ountryCode\x18\x04 \x01(\t\x12\x11\n\tdeviceUID\x18\x05 \x01(\x0c\x12\x11\n\tipAddress\x18\x06 \x01(\t\"\xe2\x05\n\x07License\x12\x18\n\x10subscriptionCode\x18\x01 \x01(\t\x12\x15\n\rproductTypeId\x18\x02 \x01(\x05\x12\x17\n\x0fproductTypeName\x18\x03 \x01(\t\x12\x16\n\x0e\x65xpirationDate\x18\x04 \x01(\t\x12\x1e\n\x16secondsUntilExpiration\x18\x05 \x01(\x03\x12\x12\n\nmaxDevices\x18\x06 \x01(\x05\x12\x14\n\x0c\x66ilePlanType\x18\x07 \x01(\x05\x12\x11\n\tbytesUsed\x18\x08 \x01(\x03\x12\x12\n\nbytesTotal\x18\t \x01(\x03\x12%\n\x1dsecondsUntilStorageExpiration\x18\n \x01(\x03\x12\x1d\n\x15storageExpirationDate\x18\x0b \x01(\t\x12,\n$hasAutoRenewableAppstoreSubscription\x18\x0c \x01(\x08\x12\x13\n\x0b\x61\x63\x63ountType\x18\r \x01(\x05\x12\x18\n\x10uploadsRemaining\x18\x0e \x01(\x05\x12\x14\n\x0c\x65nterpriseId\x18\x0f \x01(\x05\x12\x13\n\x0b\x63hatEnabled\x18\x10 \x01(\x08\x12 \n\x18\x61uditAndReportingEnabled\x18\x11 \x01(\x08\x12!\n\x19\x62reachWatchFeatureDisable\x18\x12 \x01(\x08\x12\x12\n\naccountUid\x18\x13 \x01(\x0c\x12\x1c\n\x14\x61llowPersonalLicense\x18\x14 \x01(\x08\x12\x12\n\nlicensedBy\x18\x15 \x01(\t\x12\r\n\x05\x65mail\x18\x16 \x01(\t\x12\x1a\n\x12\x62reachWatchEnabled\x18\x17 \x01(\x08\x12\x1a\n\x12\x62reachWatchScanned\x18\x18 \x01(\x08\x12\x1d\n\x15\x62reachWatchExpiration\x18\x19 \x01(\x03\x12\x1e\n\x16\x62reachWatchDateCreated\x18\x1a \x01(\x03\x12%\n\x05\x65rror\x18\x1b \x01(\x0b\x32\x16.AccountSummary.Result\"\xa3\x01\n\x05\x41\x64\x64On\x12\x14\n\x0clicenseKeyId\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x16\n\x0e\x65xpirationDate\x18\x03 \x01(\x03\x12\x13\n\x0b\x63reatedDate\x18\x04 \x01(\x03\x12\x0f\n\x07isTrial\x18\x05 \x01(\x08\x12\x0f\n\x07\x65nabled\x18\x06 \x01(\x08\x12\x0f\n\x07scanned\x18\x07 \x01(\x08\x12\x16\n\x0e\x66\x65\x61tureDisable\x18\x08 \x01(\x08\"\x8d\x08\n\x08Settings\x12\r\n\x05\x61udit\x18\x01 \x01(\x08\x12!\n\x19mustPerformAccountShareBy\x18\x02 \x01(\x03\x12>\n\x0eshareAccountTo\x18\x03 \x03(\x0b\x32&.AccountSummary.MissingAccountShareKey\x12+\n\x05rules\x18\x04 \x03(\x0b\x32\x1c.AccountSummary.PasswordRule\x12\x1a\n\x12passwordRulesIntro\x18\x05 \x01(\t\x12\x16\n\x0e\x61utoBackupDays\x18\x06 \x01(\x05\x12\r\n\x05theme\x18\x07 \x01(\t\x12\x0f\n\x07\x63hannel\x18\x08 \x01(\t\x12\x14\n\x0c\x63hannelValue\x18\t \x01(\t\x12\x15\n\rrsaConfigured\x18\n \x01(\x08\x12\x15\n\remailVerified\x18\x0b \x01(\x08\x12\"\n\x1amasterPasswordLastModified\x18\x0c \x01(\x01\x12\x18\n\x10\x61\x63\x63ountFolderKey\x18\r \x01(\x0c\x12\x31\n\x0csecurityKeys\x18\x0e \x03(\x0b\x32\x1b.AccountSummary.SecurityKey\x12+\n\tkeyValues\x18\x0f \x03(\x0b\x32\x18.AccountSummary.KeyValue\x12\x0f\n\x07ssoUser\x18\x10 \x01(\x08\x12\x18\n\x10onlineAccessOnly\x18\x11 \x01(\x08\x12\x1c\n\x14masterPasswordExpiry\x18\x12 \x01(\x05\x12\x19\n\x11twoFactorRequired\x18\x13 \x01(\x08\x12\x16\n\x0e\x64isallowExport\x18\x14 \x01(\x08\x12\x15\n\rrestrictFiles\x18\x15 \x01(\x08\x12\x1a\n\x12restrictAllSharing\x18\x16 \x01(\x08\x12\x17\n\x0frestrictSharing\x18\x17 \x01(\x08\x12\"\n\x1arestrictSharingIncomingAll\x18\x18 \x01(\x08\x12)\n!restrictSharingIncomingEnterprise\x18\x19 \x01(\x08\x12\x13\n\x0blogoutTimer\x18\x1a \x01(\x03\x12\x17\n\x0fpersistentLogin\x18\x1b \x01(\x08\x12\x1c\n\x14ipDisableAutoApprove\x18\x1c \x01(\x08\x12$\n\x1cshareDataKeyWithEccPublicKey\x18\x1d \x01(\x08\x12\'\n\x1fshareDataKeyWithDevicePublicKey\x18\x1e \x01(\x08\x12\x1a\n\x12RecordTypesCounter\x18\x1f \x01(\x05\x12$\n\x1cRecordTypesEnterpriseCounter\x18  \x01(\x05\x12\x1a\n\x12recordTypesEnabled\x18! \x01(\x08\x12\x1c\n\x14\x63\x61nManageRecordTypes\x18\" \x01(\x08\"&\n\x08KeyValue\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"-\n\x0fKeyValueBoolean\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08\"*\n\x0cKeyValueLong\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x03\"=\n\x06Result\x12\x12\n\nresultCode\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x0e\n\x06result\x18\x03 \x01(\t\"\xc2\x01\n\x0c\x45nforcements\x12)\n\x07strings\x18\x01 \x03(\x0b\x32\x18.AccountSummary.KeyValue\x12\x31\n\x08\x62ooleans\x18\x02 \x03(\x0b\x32\x1f.AccountSummary.KeyValueBoolean\x12+\n\x05longs\x18\x03 \x03(\x0b\x32\x1c.AccountSummary.KeyValueLong\x12\'\n\x05jsons\x18\x04 \x03(\x0b\x32\x18.AccountSummary.KeyValue\"<\n\x16MissingAccountShareKey\x12\x0f\n\x07role_id\x18\x01 \x01(\x03\x12\x11\n\tpublicKey\x18\x02 \x01(\x0c\"u\n\x0cPasswordRule\x12\x10\n\x08ruleType\x18\x01 \x01(\t\x12\x0f\n\x07pattern\x18\x02 \x01(\t\x12\r\n\x05match\x18\x03 \x01(\x08\x12\x0f\n\x07minimum\x18\x04 \x01(\x05\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\r\n\x05value\x18\x06 \x01(\t\"\x97\x01\n\x0bSecurityKey\x12\x10\n\x08\x64\x65viceId\x18\x01 \x01(\x03\x12\x12\n\ndeviceName\x18\x02 \x01(\t\x12\x11\n\tdateAdded\x18\x03 \x01(\x03\x12\x0f\n\x07isValid\x18\x04 \x01(\x08\x12>\n\x12\x64\x65viceRegistration\x18\x05 \x01(\x0b\x32\".AccountSummary.DeviceRegistration\"y\n\x12\x44\x65viceRegistration\x12\x11\n\tkeyHandle\x18\x01 \x01(\t\x12\x11\n\tpublicKey\x18\x02 \x01(\x0c\x12\x17\n\x0f\x61ttestationCert\x18\x03 \x01(\t\x12\x0f\n\x07\x63ounter\x18\x04 \x01(\x03\x12\x13\n\x0b\x63ompromised\x18\x05 \x01(\x08\"k\n\x05Group\x12\r\n\x05\x61\x64min\x18\x01 \x01(\x08\x12\x1d\n\x15groupVerificationCode\x18\x02 \x01(\t\x12\x34\n\radministrator\x18\x04 \x01(\x0b\x32\x1d.AccountSummary.Administrator\"\xc0\x01\n\rAdministrator\x12\x11\n\tfirstName\x18\x01 \x01(\t\x12\x10\n\x08lastName\x18\x02 \x01(\t\x12\r\n\x05\x65mail\x18\x03 \x01(\t\x12\x1c\n\x14\x63urrentNumberOfUsers\x18\x04 \x01(\x05\x12\x15\n\rnumberOfUsers\x18\x05 \x01(\x05\x12\x18\n\x10subscriptionCode\x18\x07 \x01(\t\x12\x16\n\x0e\x65xpirationDate\x18\x08 \x01(\t\x12\x14\n\x0cpurchaseDate\x18\t \x01(\tB*\n\x18\x63om.keepersecurity.protoB\x0e\x41\x63\x63ountSummaryb\x06proto3')



_ACCOUNTSUMMARYREQUEST = DESCRIPTOR.message_types_by_name['AccountSummaryRequest']
_ACCOUNTSUMMARYELEMENTS = DESCRIPTOR.message_types_by_name['AccountSummaryElements']
_DEVICEINFO = DESCRIPTOR.message_types_by_name['DeviceInfo']
_KEYSINFO = DESCRIPTOR.message_types_by_name['KeysInfo']
_SYNCLOG = DESCRIPTOR.message_types_by_name['SyncLog']
_LICENSE = DESCRIPTOR.message_types_by_name['License']
_ADDON = DESCRIPTOR.message_types_by_name['AddOn']
_SETTINGS = DESCRIPTOR.message_types_by_name['Settings']
_KEYVALUE = DESCRIPTOR.message_types_by_name['KeyValue']
_KEYVALUEBOOLEAN = DESCRIPTOR.message_types_by_name['KeyValueBoolean']
_KEYVALUELONG = DESCRIPTOR.message_types_by_name['KeyValueLong']
_RESULT = DESCRIPTOR.message_types_by_name['Result']
_ENFORCEMENTS = DESCRIPTOR.message_types_by_name['Enforcements']
_MISSINGACCOUNTSHAREKEY = DESCRIPTOR.message_types_by_name['MissingAccountShareKey']
_PASSWORDRULE = DESCRIPTOR.message_types_by_name['PasswordRule']
_SECURITYKEY = DESCRIPTOR.message_types_by_name['SecurityKey']
_DEVICEREGISTRATION = DESCRIPTOR.message_types_by_name['DeviceRegistration']
_GROUP = DESCRIPTOR.message_types_by_name['Group']
_ADMINISTRATOR = DESCRIPTOR.message_types_by_name['Administrator']
AccountSummaryRequest = _reflection.GeneratedProtocolMessageType('AccountSummaryRequest', (_message.Message,), {
  'DESCRIPTOR' : _ACCOUNTSUMMARYREQUEST,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.AccountSummaryRequest)
  })
_sym_db.RegisterMessage(AccountSummaryRequest)

AccountSummaryElements = _reflection.GeneratedProtocolMessageType('AccountSummaryElements', (_message.Message,), {
  'DESCRIPTOR' : _ACCOUNTSUMMARYELEMENTS,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.AccountSummaryElements)
  })
_sym_db.RegisterMessage(AccountSummaryElements)

DeviceInfo = _reflection.GeneratedProtocolMessageType('DeviceInfo', (_message.Message,), {
  'DESCRIPTOR' : _DEVICEINFO,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.DeviceInfo)
  })
_sym_db.RegisterMessage(DeviceInfo)

KeysInfo = _reflection.GeneratedProtocolMessageType('KeysInfo', (_message.Message,), {
  'DESCRIPTOR' : _KEYSINFO,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.KeysInfo)
  })
_sym_db.RegisterMessage(KeysInfo)

SyncLog = _reflection.GeneratedProtocolMessageType('SyncLog', (_message.Message,), {
  'DESCRIPTOR' : _SYNCLOG,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.SyncLog)
  })
_sym_db.RegisterMessage(SyncLog)

License = _reflection.GeneratedProtocolMessageType('License', (_message.Message,), {
  'DESCRIPTOR' : _LICENSE,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.License)
  })
_sym_db.RegisterMessage(License)

AddOn = _reflection.GeneratedProtocolMessageType('AddOn', (_message.Message,), {
  'DESCRIPTOR' : _ADDON,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.AddOn)
  })
_sym_db.RegisterMessage(AddOn)

Settings = _reflection.GeneratedProtocolMessageType('Settings', (_message.Message,), {
  'DESCRIPTOR' : _SETTINGS,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.Settings)
  })
_sym_db.RegisterMessage(Settings)

KeyValue = _reflection.GeneratedProtocolMessageType('KeyValue', (_message.Message,), {
  'DESCRIPTOR' : _KEYVALUE,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.KeyValue)
  })
_sym_db.RegisterMessage(KeyValue)

KeyValueBoolean = _reflection.GeneratedProtocolMessageType('KeyValueBoolean', (_message.Message,), {
  'DESCRIPTOR' : _KEYVALUEBOOLEAN,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.KeyValueBoolean)
  })
_sym_db.RegisterMessage(KeyValueBoolean)

KeyValueLong = _reflection.GeneratedProtocolMessageType('KeyValueLong', (_message.Message,), {
  'DESCRIPTOR' : _KEYVALUELONG,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.KeyValueLong)
  })
_sym_db.RegisterMessage(KeyValueLong)

Result = _reflection.GeneratedProtocolMessageType('Result', (_message.Message,), {
  'DESCRIPTOR' : _RESULT,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.Result)
  })
_sym_db.RegisterMessage(Result)

Enforcements = _reflection.GeneratedProtocolMessageType('Enforcements', (_message.Message,), {
  'DESCRIPTOR' : _ENFORCEMENTS,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.Enforcements)
  })
_sym_db.RegisterMessage(Enforcements)

MissingAccountShareKey = _reflection.GeneratedProtocolMessageType('MissingAccountShareKey', (_message.Message,), {
  'DESCRIPTOR' : _MISSINGACCOUNTSHAREKEY,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.MissingAccountShareKey)
  })
_sym_db.RegisterMessage(MissingAccountShareKey)

PasswordRule = _reflection.GeneratedProtocolMessageType('PasswordRule', (_message.Message,), {
  'DESCRIPTOR' : _PASSWORDRULE,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.PasswordRule)
  })
_sym_db.RegisterMessage(PasswordRule)

SecurityKey = _reflection.GeneratedProtocolMessageType('SecurityKey', (_message.Message,), {
  'DESCRIPTOR' : _SECURITYKEY,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.SecurityKey)
  })
_sym_db.RegisterMessage(SecurityKey)

DeviceRegistration = _reflection.GeneratedProtocolMessageType('DeviceRegistration', (_message.Message,), {
  'DESCRIPTOR' : _DEVICEREGISTRATION,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.DeviceRegistration)
  })
_sym_db.RegisterMessage(DeviceRegistration)

Group = _reflection.GeneratedProtocolMessageType('Group', (_message.Message,), {
  'DESCRIPTOR' : _GROUP,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.Group)
  })
_sym_db.RegisterMessage(Group)

Administrator = _reflection.GeneratedProtocolMessageType('Administrator', (_message.Message,), {
  'DESCRIPTOR' : _ADMINISTRATOR,
  '__module__' : 'AccountSummary_pb2'
  # @@protoc_insertion_point(class_scope:AccountSummary.Administrator)
  })
_sym_db.RegisterMessage(Administrator)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030com.keepersecurity.protoB\016AccountSummary'
  _ACCOUNTSUMMARYREQUEST._serialized_start=58
  _ACCOUNTSUMMARYREQUEST._serialized_end=105
  _ACCOUNTSUMMARYELEMENTS._serialized_start=108
  _ACCOUNTSUMMARYELEMENTS._serialized_end=651
  _DEVICEINFO._serialized_start=654
  _DEVICEINFO._serialized_end=965
  _KEYSINFO._serialized_start=968
  _KEYSINFO._serialized_end=1161
  _SYNCLOG._serialized_start=1164
  _SYNCLOG._serialized_end=1293
  _LICENSE._serialized_start=1296
  _LICENSE._serialized_end=2034
  _ADDON._serialized_start=2037
  _ADDON._serialized_end=2200
  _SETTINGS._serialized_start=2203
  _SETTINGS._serialized_end=3240
  _KEYVALUE._serialized_start=3242
  _KEYVALUE._serialized_end=3280
  _KEYVALUEBOOLEAN._serialized_start=3282
  _KEYVALUEBOOLEAN._serialized_end=3327
  _KEYVALUELONG._serialized_start=3329
  _KEYVALUELONG._serialized_end=3371
  _RESULT._serialized_start=3373
  _RESULT._serialized_end=3434
  _ENFORCEMENTS._serialized_start=3437
  _ENFORCEMENTS._serialized_end=3631
  _MISSINGACCOUNTSHAREKEY._serialized_start=3633
  _MISSINGACCOUNTSHAREKEY._serialized_end=3693
  _PASSWORDRULE._serialized_start=3695
  _PASSWORDRULE._serialized_end=3812
  _SECURITYKEY._serialized_start=3815
  _SECURITYKEY._serialized_end=3966
  _DEVICEREGISTRATION._serialized_start=3968
  _DEVICEREGISTRATION._serialized_end=4089
  _GROUP._serialized_start=4091
  _GROUP._serialized_end=4198
  _ADMINISTRATOR._serialized_start=4201
  _ADMINISTRATOR._serialized_end=4393
# @@protoc_insertion_point(module_scope)



import logging
from datetime import timedelta
from re import findall

from ...constants import (
    TIMEOUT_DEFAULT, TIMEOUT_MIN, TIMEOUT_DEFAULT_UNIT, TIMEOUT_ALLOWED_UNITS
)


def parse_timeout(timeout_input):
    """Parse timeout input to return instance of timedelta

    timeout_input(str): String to parse for one or more integer values followed by time unit names. The allowed time
        units are found in the list constants.TIMEOUT_ALLOWED_UNITS. Any substring from the beginning of the allowed
        unit may be provided. If no time unit names are included, then the default time unit from
        constants.TIMEOUT_DEFAULT_UNIT is used. An error is raised if a unit name in timeout_input is unrecognized.
    Returns instance of timedelta from the unit values and names.
    """
    if timeout_input.strip().isnumeric():
        tdelta_kwargs = {TIMEOUT_DEFAULT_UNIT: int(timeout_input)}
    else:
        all_units = TIMEOUT_ALLOWED_UNITS
        tdelta_kwargs = {}
        for v, input_unit in findall(r'(\d+)\s*([a-zA-Z]+)\s*', timeout_input):
            key_match = [t for t in all_units if t.startswith(input_unit)]
            if len(key_match) == 0:
                raise ValueError(
                    f'{input_unit} is not allowed as a unit for the timeout value. '
                    f'Valid units for the timeout value are {TIMEOUT_ALLOWED_UNITS}.'
                )
            tdelta_kwargs[key_match[0]] = int(v)
    return timedelta(**tdelta_kwargs)


def format_timeout(timeout_delta):
    """Format timeout as instance of timedelta for output to console

    timeout_delta(timedelta): Timeout setting as instance of timedelta.
    Returns string with value and names of time units separated by commas. If all unit values are zero,
        then the string '0' is returned.
    """
    if timeout_delta == timedelta(0):
        return '0'
    else:
        time_units = {'days': timeout_delta.days}
        hours_minutes = str(timeout_delta).split(', ')[-1]
        time_units['hours'], time_units['minutes'], _ = hours_minutes.split(':')
        for k, v in time_units.items():
            time_units[k] = int(v)
        nonzero_units = [f'{v} {k[:-1] if v == 1 else k}' for k, v in time_units.items() if v != 0]
        return ', '.join(nonzero_units)


def enforce_timeout_range(timeout_delta):
    """Enforce the range of allowed timeout values based on constants TIMEOUT_MIN
    Warnings are raised if the timeout is outside of the range.

    timeout_delta(timedelta): Timeout setting as instance of timedelta.
    """
    if timeout_delta < TIMEOUT_MIN:
        logging.warning(
            f'The minimum device timeout value is {format_timeout(TIMEOUT_MIN)}. '
            'The device timeout has been set to the default Keeper timeout value.'
        )
        return TIMEOUT_DEFAULT
    else:
        return timeout_delta


def get_timeout_setting_from_delta(timeout_delta):
    """Get timeout setting in minutes to be used in an API call to the Keeper backend

    timeout_delta(timedelta): Timeout setting as instance of timedelta.
    Returns string representation of the integer value of timeout minutes.
    """
    timeout_seconds = int(timeout_delta.total_seconds() // 60)
    return str(timeout_seconds)


def get_delta_from_timeout_setting(timeout_setting):
    """Get the timeout as an instance of timedelta from the setting returned from the API call

    timeout_setting(str): Timeout setting as a string representation of the integer value of timeout milliseconds.
    Returns timedelta instance.
    """
    return timedelta(milliseconds=int(timeout_setting))



# coding: utf-8
class Account(object):
    def __init__(self, id, name, username, password, url, group, notes=None, shared_folder=None, attach_key=None,
                 totp_secret=None, totp_url=None):
        self.id = id
        self.name = name
        self.username = username
        self.password = password
        self.url = url
        self.group = group
        self.notes = notes
        self.shared_folder = shared_folder
        self.attach_key = attach_key
        self.totp_secret = totp_secret
        self.totp_url = totp_url
        self.attachments = []



import io
from unittest import TestCase

from keepercommander import crypto, utils


class TestCrypto(TestCase):

    def test_stream_decrypter(self):
        data = bytearray(999)
        for i in range(len(data)):
            data[i] = i & 0xff

        key = utils.generate_aes_key()

        crypter = crypto.StreamCrypter()
        crypter.key = key
        crypter.is_gcm = False

        encrypted_data = bytearray()
        with crypter.set_stream(io.BytesIO(data), True) as cs:
            while True:
                buffer = cs.read(1024)
                if not buffer:
                    break
                encrypted_data.extend(buffer)
        self.assertEqual(crypter.bytes_read, len(data))

        decrypted_data = bytearray()
        with crypter.set_stream(io.BytesIO(encrypted_data), False) as cs:
            while True:
                buffer = cs.read(1024)
                if not buffer:
                    break
                decrypted_data.extend(buffer)

        self.assertEqual(decrypted_data, data)

    def test_decrypt_aes_v1(self):
        data = utils.base64_url_decode('KvsOJmE4JNK1HwKSpkBeR5R9YDms86uOb3wjNvc4LbUnZhKQtDxWifgA99tH2ZuP')
        key = utils.base64_url_decode('pAZmcxEoV2chXsFQ6bzn7Lop8yO4F8ERIuS7XpFtr7Y')
        data = crypto.decrypt_aes_v1(data, key)
        self.assertEqual(data, utils.base64_url_decode('6lf4FGVyhDRnRhJ91TrahjIW8lTqGA'))

    def test_encrypt_aes_v2(self):
        key = utils.base64_url_decode('c-EeCGlAO7F9QoJThlFBrhSCLYMe1H6GtKP-rezDnik')
        data = utils.base64_url_decode('nm-8mRG7xYwUG2duaOZzw-ttuqfetWjVIzoridJF0EJOGlDLs1ZWQ7F9mOJ0Hxuy' +
                                       'dFyojxdxVo1fGwbfwf0Jew07HhGGE5UZ_s57rQvhizDW3F3z9a7EqHRon0EilCbMhIzE')
        nonce = utils.base64_url_decode('Nt9_Y37C_43eRCRQ')
        enc_data = crypto.encrypt_aes_v2(data, key, nonce)
        expected_data = utils.base64_url_decode('Nt9_Y37C_43eRCRQCptb64zFaJVLcXF1udabOr_fyGXkpjpYeCAI7zVQD4JjewB' +
                                                'CP1Xp7D6dx-pxdRWkhDEnVhJ3fzezi8atmmzvf2ICfkDK0IHHB8iNSx_R1Ru8Tozb-IdavT3wKi7nKSJLDdt-dk-Mw7bCewpZtg4wY-1UQw')
        self.assertEqual(enc_data, expected_data)

        dec_data = crypto.decrypt_aes_v2(enc_data, key)
        self.assertEqual(dec_data, data)

    def test_encrypt_aes_v1(self):
        iv = utils.base64_url_decode('KvsOJmE4JNK1HwKSpkBeRw')
        block = utils.base64_url_decode('6lf4FGVyhDRnRhJ91TrahjIW8lTqGA')
        key = utils.base64_url_decode('pAZmcxEoV2chXsFQ6bzn7Lop8yO4F8ERIuS7XpFtr7Y')
        enc = crypto.encrypt_aes_v1(block, key, iv)
        encoded = utils.base64_url_encode(enc)
        self.assertEqual(encoded, 'KvsOJmE4JNK1HwKSpkBeR5R9YDms86uOb3wjNvc4LbUnZhKQtDxWifgA99tH2ZuP')

    def test_encrypt_rsa(self):
        data = crypto.get_random_bytes(100)
        puk = crypto.load_rsa_public_key(utils.base64_url_decode(_test_public_key))
        enc_data = crypto.encrypt_rsa(data, puk)
        prk = crypto.load_rsa_private_key(utils.base64_url_decode(_test_private_key))
        dec_data = crypto.decrypt_rsa(enc_data, prk)
        self.assertEqual(data, dec_data)

    def test_ec_encryption(self):
        private_key = crypto.load_ec_private_key(utils.base64_url_decode(_test_ec_private_key))
        decrypted_data = crypto.decrypt_ec(utils.base64_url_decode(_test_ec_encrypted_data), private_key)
        data = utils.base64_url_decode(_test_random_data)
        self.assertEqual(decrypted_data, data)

    def test_derive_key_hash_v1(self):
        password = 'q2rXmNBFeLwAEX55hVVTfg'
        salt = utils.base64_url_decode('Ozv5_XSBgw-XSrDosp8Y1A')
        iterations = 1000
        expected_key = utils.base64_url_decode('nu911pKhOIeX_lToXa4uIUuMPg1pj_3ZGpGmd7OjvRs')
        key_hash = crypto.derive_keyhash_v1(password, salt, iterations)
        self.assertEqual(key_hash, expected_key)

    def test_derive_key_hash_v2(self):
        password = 'q2rXmNBFeLwAEX55hVVTfg'
        salt = utils.base64_url_decode('Ozv5_XSBgw-XSrDosp8Y1A')
        iterations = 1000
        expected_key = utils.base64_url_decode('rXE9OHv_gcvUHdWuBIkyLsRDXT1oddQCzf6PrIECl2g')
        domain = '1oZZl0fKjU4'
        key_hash = crypto.derive_keyhash_v2(domain, password, salt, iterations)
        self.assertEqual(key_hash, expected_key)

    def test_password_score(self):
        self.assertEqual(utils.password_score('!@#$%^&*()'), 92)
        self.assertEqual(utils.password_score('aZkljfzsnmp4w9058dsqln5yf(&*))(*)(345'), 100)
        self.assertEqual(utils.password_score('c3>^sxuKZ[Ndyo(OBE14'), 100)
        self.assertEqual(utils.password_score('AAAbbbCCC11'), 38)
        self.assertEqual(utils.password_score('password'), 8)


_test_random_data = \
    'cKGoVph_X0NKjk8jQgxyQWRElUY7IsbbIJaRcJVlnOb7AchFiY-izmTTOlgArwIqAxKDKSRAWx2Q1pX' \
    'mWUUoEVKwFE9B2pLUi_GrTQu_hPEm6HjinwJaUPql-kqDfvGLiX6yBw'

_test_ec_private_key = 'HIIeyuuRkVGvhtax8mlX7fangaC6DKa2R8VAg5AAtBY'

_test_ec_encrypted_data = \
    'BMObx0eJO48N3Fg-2rQzVzppGi51kkxCM3KTRr2Zqe_q1kS8waisb2X9mDMnGKCtJqJLsVdTYsCn_BZ' \
    'zrbXO8846fMBqc1fKnOIBnxc_YIvTU9iKdx3vKgonOrPkyLMu273YQx8MpP-_FUjymRG7Rf4J0BIroP' \
    '5ndvaC6bzDmZGS-KLP8zO7oY7oNgeOI9Tj3AXCW65No2QzvGsppEHK1DcPpL8dh4RK2vt8Hh9Udj86Y' \
    'd-pi11AXusnzysl6cwzWA'

_test_public_key = "MIIBCgKCAQEAqR0AjmBXo371pYmvS1NM8nXlbAv5qUbPYuV6KVwKjN3T8WX5K6HD" \
                   "Gl3-ylAbI02vIzKue-gDbjo1wUGp2qhANc1VxllLSWnkJmwbuGUTEWp4ANjusoMh" \
                   "PvEwna1XPdlrSMdsKokjbP9xbguPdvXx5oBaqArrrGEg-36Vi7miA_g_UT4DKcry" \
                   "glD4Xx0H9t5Hav-frz2qcEsyh9FC0fNyon_uveEdP2ac-kax8vO5EeVfBzOdw-WP" \
                   "aBtUO1h7rSZ6xKOm6x1OahNTUFy7Cgm038JuMwHChTK29H9EOlqbOOuzYA1ENzL8" \
                   "8hELpe-kl4RmpNS94BJDssikFFbjoiAVfwIDAQAB"

_test_private_key = "MIIEogIBAAKCAQEAqR0AjmBXo371pYmvS1NM8nXlbAv5qUbPYuV6KVwKjN3T8WX5" \
                    "K6HDGl3-ylAbI02vIzKue-gDbjo1wUGp2qhANc1VxllLSWnkJmwbuGUTEWp4ANju" \
                    "soMhPvEwna1XPdlrSMdsKokjbP9xbguPdvXx5oBaqArrrGEg-36Vi7miA_g_UT4D" \
                    "KcryglD4Xx0H9t5Hav-frz2qcEsyh9FC0fNyon_uveEdP2ac-kax8vO5EeVfBzOd" \
                    "w-WPaBtUO1h7rSZ6xKOm6x1OahNTUFy7Cgm038JuMwHChTK29H9EOlqbOOuzYA1E" \
                    "NzL88hELpe-kl4RmpNS94BJDssikFFbjoiAVfwIDAQABAoIBABB9KW64ahMg7-ai" \
                    "FBtuFdSWNjZgvIkKxHHKGi0qMkUl4-JnpPHiJdnOTGeBhAPfMTJnYKfoKV14A4HC" \
                    "W0NcoFYenTxnvHV-A6bTZ6iFAmTyUp0SicOSEY3Hiov1OMppBpLkDuHe2TtpdK_c" \
                    "JLLerCVjYnN8DRqTpdmfsAkdonRseXyhRhwO6yFwVy9TEc9_OFuqGMOsy5_VIts6" \
                    "pG0saJJUQlOuLTxHwtPdloqjI8l3yMiDfXvJF2_epb_PYpKkAQZy_UWM5u4P_pnb" \
                    "UdImyYo6HBmnq-qO07J7b3yOSAzWhklBD7cMh1ucSOyF9-u03mLOfx2-SXq4tIuU" \
                    "Lz3RHZECgYEA0Rj-ipCKEPwQORViDFYYk1txzFSVKVX9Q-ozl6i93kTXx8GF7vkX" \
                    "L6SaEbKDA2EARuczr1gjymlvgRAwbsX7bDylSF6EsmPZ-EccNe4GoXmfbgMFDqGr" \
                    "3jVUmwEYwkte6EvP2Ha2GDwIuXFhcXWxgbbQxGGEcS5niei1mV0jv-sCgYEAzwv9" \
                    "BIYkeBC6_kejD2VwNzC1Jl97vg2It2URTZUGPFvcXh1Ed_i1itXwJ7wBjyBdwLJM" \
                    "IWjZcAYKET9NdBps2loATbOHrw4zFEqjKr_X-xSVU4bunipoY40fhl6a15ngUZ49" \
                    "3OJe_YtXEBHTVHorltIYuugu0zKk6uKbU_bt770CgYAR8_5u8UgZezr9W7umaYIE" \
                    "rPZRX_XKrcpoGWTCocdjnS-VxCT2xsZZ3d0opdYf5SU78T_7zyqLh4_-WeB-slsL" \
                    "CQ3777mfA3nEmn5ulvhUxveMX5AAmJsEIjoYcPiqPgRxF4lKAa9S11y8Z2LBdiR-" \
                    "ia7VHbZcbWqQab2l5FxcbwKBgCz_Ov7XtGdPo4QNx5daAVhNQqFTUQ5N3K-WzHri" \
                    "71cA09S0YaP9Ll88_ZN1HZWggB-X4EnGgrMA7QEwk8Gu2Idf1f8NDGj0Gg_H5Mwu" \
                    "o17S610azxMavlMcYYSPXPGMZJ74WBOAMwrBVKuOZDJQ1tZRVMSSH1MRB5xwoTdP" \
                    "TAi1AoGAXqJUfDAjtLR0wFoLlV0GWGOObKkPZFCbFdv0_CY2dk0nKnSsYRCogiFP" \
                    "t9XhZG5kawEtdfqiNBDyeNVLu6FaZnRkid_tUqMKfCYLjNDq31OD1Pwvyuh6Hs1P" \
                    "hL2-nt6t9b7JMyzKjWq_OPuTPH0QErL3oiFbTaZ4fDXplH_6Snw"



#  _  __  
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|            
#
# Keeper Commander 
# Contact: ops@keepersecurity.com
#
import json
import shutil
from collections import OrderedDict as OD
from typing import Tuple, List, Union

from asciitree import LeftAligned
from colorama import init, Fore, Back, Style
from tabulate import tabulate

from keepercommander import __version__
from .subfolder import BaseFolderNode, SharedFolderNode

init()


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    HIGHINTENSITYRED = '\033[1;91m'


def welcome():
    lines = []    # type: List[Union[str, Tuple[str, str]]]

    lines.append( r'         /#############/   /#\ ')
    lines.append( r'        /#############/   /###\      _    __  _______  _______  ______   _______  ______ (R)')
    lines.append( r'       /#############/   /#####\    | |  / / |  ____/ |  ____/ |  ___ \ |  ____/ |  ___ \ ')
    lines.append( r'      /######/           \######\   | | / /  | |____  | |____  | | __| || |____  | | __| | ')
    lines.append( r'     /######/             \######\  | |< <   |  ___/  |  ___/  | |/___/ |  ___/  | |/_  / ')
    lines.append( r'    /######/               \######\ | | \ \  | |_____ | |_____ | |      | |_____ | |  \ \ ')
    lines.append( r'    \######\               /######/ |_|  \_\ |_______||_______||_|      |_______||_|   \_\ ')
    lines.append((r'     \######\             /######/', r'     ____                                          _ '))
    lines.append((r'      \######\           /######/ ', r'   /  ___|___  _ __ ___  _ __ ___   __ _ _ __   __| | ___ _ __ '))
    lines.append((r'       \#############\   \#####/  ', r"  /  /   / _ \| '_ ` _ \| '_ ` _ \ / _` | '_ \ / _` |/ _ \ '__| "))
    lines.append((r'        \#############\   \###/   ', r'  \  \__| (_) | | | | | | | | | | | (_| | | | | (_| |  __/ | '))
    lines.append((r'         \#############\   \#/    ', r'   \_____\___/|_| |_| |_|_| |_| |_|\__,_|_| |_|\__,_|\___|_| '))
    lines.append('')

    try:
        width = shutil.get_terminal_size(fallback=(160, 50)).columns
    except:
        width = 160
    print(Style.RESET_ALL)
    print(Back.BLACK + Style.BRIGHT + '\n')
    for line in lines:
        if isinstance(line, str):
            if len(line) > width:
                line = line[:width]
            print('\033[2K' + Fore.LIGHTYELLOW_EX + line)
        elif isinstance(line, tuple):
            yellow_line = line[0] if len(line) > 0 else ''
            white_line = line[1] if len(line) > 1 else ''
            if len(yellow_line) > width:
                yellow_line = yellow_line[:width]
            if len(yellow_line) + len(white_line) > width:
                if len(yellow_line) < width:
                    white_line = white_line[:width - len(yellow_line)]
                else:
                    white_line = ''
            print('\033[2K' + Fore.LIGHTYELLOW_EX + yellow_line + Fore.LIGHTWHITE_EX + white_line)

    print('\033[2K' + Fore.LIGHTBLACK_EX + f'{("v" + __version__):>93}\n' + Style.RESET_ALL)


def formatted_records(records, **kwargs):
    """Display folders/titles/uids for the supplied shared folders"""
    params = None
    if 'params' in kwargs:
        params = kwargs['params']

    # Sort by title
    records.sort(key=lambda x: x.title.lower(), reverse=False)

    def abbreviate_text(text: str, chars_num: int):
        if 'verbose' in kwargs and kwargs['verbose']:
            return text
        else:
            return text if len(text) < chars_num else text[:chars_num] + '...'

    if len(records) > 0:
        shared_folder = None
        if 'folder' in kwargs and params is not None:
            fuid = kwargs['folder']
            if fuid in params.folder_cache:
                folder = params.folder_cache[fuid]
                if folder.type in {BaseFolderNode.SharedFolderType, BaseFolderNode.SharedFolderFolderType}:
                    if folder.type == BaseFolderNode.SharedFolderFolderType:
                        fuid = folder.shared_folder_uid
                else:
                    fuid = None
                if fuid and fuid in params.shared_folder_cache:
                    shared_folder = params.shared_folder_cache[fuid]

        table = [[i + 1, r.record_uid, abbreviate_text(r.record_type, 32), abbreviate_text(r.title, 32), r.login, abbreviate_text(r.login_url, 32)] for i, r in enumerate(records)]
        headers = ["#", 'Record UID', 'Type', 'Title', 'Login', 'URL']
        if shared_folder and 'records' in shared_folder:
            headers.append('Permissions')
            for row in table:
                permissions = ''
                for sfr in shared_folder['records']:
                    if sfr['record_uid'] == row[1]:
                        if sfr['can_edit']:
                            permissions = 'Can Edit'
                        if sfr['can_share']:
                            if permissions:
                                permissions += ' & Share'
                            else:
                                permissions = 'Can Share'
                        if not permissions:
                            permissions = 'Read Only'
                        break
                row.append(permissions)

        print(tabulate(table, headers=headers))


def formatted_shared_folders(shared_folders, **kwargs):
    """Display folders/titles/uids for the supplied records"""

    # Sort by folder+title
    shared_folders.sort(key=lambda x: (x.name if x.name else ' ').lower(), reverse=False)

    if len(shared_folders) > 0:

        table = [[i + 1, sf.shared_folder_uid, sf.name] for i, sf in enumerate(shared_folders)]
        print(tabulate(table, headers=["#", 'Shared Folder UID', 'Name']))

        print('')

    skip_details = kwargs.get('skip_details') or False
    # Under 5 recs, just display on the screen
    if len(shared_folders) < 5 and not skip_details:
        for sf in shared_folders:
            sf.display()


def formatted_teams(teams, **kwargs):
    """Display names/uids for the supplied teams"""

    # Sort by name
    teams.sort(key=lambda x: (x.name if x.name else ' ').lower(), reverse=False)

    if len(teams) > 0:

        table = [[i + 1, team.team_uid, team.name] for i, team in enumerate(teams)]
        print(tabulate(table, headers=["#", 'Team UID', 'Name']))

        print('')

    skip_details = kwargs.get('skip_details') or False
    # Under 5 recs, just display on the screen
    if len(teams) < 5 and not skip_details:
        for team in teams:
            team.display()


def formatted_folders(folders):
    def folder_flags(f):
        flags = ''
        if f.type == 'shared_folder':
            flags = flags + 'S'
        return flags

    if len(folders) > 0:
        folders.sort(key=lambda x: (x.name or ' ').lower(), reverse=False)

        table = [[i + 1, f.uid, f.name, folder_flags(f)] for i, f in enumerate(folders)]
        print(tabulate(table, headers=["#", 'Folder UID', 'Name', 'Flags']))
        print('')


def formatted_tree(params, folder, verbose=False):
    def tree_node(node):
        if verbose and node.uid:
            name = f'{node.name} ({node.uid})'
        else:
            name = node.name

        if isinstance(node, SharedFolderNode):
            name += ' ' + Style.BRIGHT + '[Shared]' + Style.NORMAL

        sfs = [params.folder_cache[sfuid] for sfuid in node.subfolders]

        if len(sfs) == 0:
            return name, {}

        sfs.sort(key=lambda f: f.name.lower(), reverse=False)
        tns = [tree_node(sf) for sf in sfs]
        return name, OD(tns)

    t = tree_node(folder)
    tree = {
        t[0]: t[1]
    }
    tr = LeftAligned()
    print(tr(tree))
    print('')


def formatted_history(history):
    """ Show the history of commands"""

    if not history: return
    if len(history) == 0: return

    print('')
    print('Command history:')
    print('----------------')

    for h in history:
        print(h)

    print('')


def print_record(params, record_uid):
    """ Show record content """

    try:
        cached_rec = params.record_cache[record_uid]
    except KeyError as e:
        raise Exception('Record not found: ' + record_uid)
    data = json.loads(cached_rec['data_unencrypted'].decode('utf-8'))
    print(data)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: SyncDown.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import record_pb2 as record__pb2
from . import breachwatch_pb2 as breachwatch__pb2
from . import APIRequest_pb2 as APIRequest__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0eSyncDown.proto\x12\x05Vault\x1a\x0crecord.proto\x1a\x11\x62reachwatch.proto\x1a\x10\x41PIRequest.proto\",\n\x0fSyncDownRequest\x12\x19\n\x11\x63ontinuationToken\x18\x01 \x01(\x0c\"\x89\x10\n\x10SyncDownResponse\x12\x19\n\x11\x63ontinuationToken\x18\x01 \x01(\x0c\x12\x0f\n\x07hasMore\x18\x02 \x01(\x08\x12\'\n\x0b\x63\x61\x63heStatus\x18\x03 \x01(\x0e\x32\x12.Vault.CacheStatus\x12&\n\x0buserFolders\x18\x04 \x03(\x0b\x32\x11.Vault.UserFolder\x12*\n\rsharedFolders\x18\x05 \x03(\x0b\x32\x13.Vault.SharedFolder\x12>\n\x17userFolderSharedFolders\x18\x06 \x03(\x0b\x32\x1d.Vault.UserFolderSharedFolder\x12\x36\n\x13sharedFolderFolders\x18\x07 \x03(\x0b\x32\x19.Vault.SharedFolderFolder\x12\x1e\n\x07records\x18\x08 \x03(\x0b\x32\r.Vault.Record\x12-\n\x0erecordMetaData\x18\t \x03(\x0b\x32\x15.Vault.RecordMetaData\x12+\n\rnonSharedData\x18\n \x03(\x0b\x32\x14.Vault.NonSharedData\x12&\n\x0brecordLinks\x18\x0b \x03(\x0b\x32\x11.Vault.RecordLink\x12\x32\n\x11userFolderRecords\x18\x0c \x03(\x0b\x32\x17.Vault.UserFolderRecord\x12\x36\n\x13sharedFolderRecords\x18\r \x03(\x0b\x32\x19.Vault.SharedFolderRecord\x12\x42\n\x19sharedFolderFolderRecords\x18\x0e \x03(\x0b\x32\x1f.Vault.SharedFolderFolderRecord\x12\x32\n\x11sharedFolderUsers\x18\x0f \x03(\x0b\x32\x17.Vault.SharedFolderUser\x12\x32\n\x11sharedFolderTeams\x18\x10 \x03(\x0b\x32\x17.Vault.SharedFolderTeam\x12\x1a\n\x12recordAddAuditData\x18\x11 \x03(\x0c\x12\x1a\n\x05teams\x18\x12 \x03(\x0b\x32\x0b.Vault.Team\x12,\n\x0esharingChanges\x18\x13 \x03(\x0b\x32\x14.Vault.SharingChange\x12\x1f\n\x07profile\x18\x14 \x01(\x0b\x32\x0e.Vault.Profile\x12%\n\nprofilePic\x18\x15 \x01(\x0b\x32\x11.Vault.ProfilePic\x12\x34\n\x12pendingTeamMembers\x18\x16 \x03(\x0b\x32\x18.Vault.PendingTeamMember\x12\x34\n\x12\x62reachWatchRecords\x18\x17 \x03(\x0b\x32\x18.Vault.BreachWatchRecord\x12\"\n\tuserAuths\x18\x18 \x03(\x0b\x32\x0f.Vault.UserAuth\x12?\n\x17\x62reachWatchSecurityData\x18\x19 \x03(\x0b\x32\x1e.Vault.BreachWatchSecurityData\x12/\n\x0freusedPasswords\x18\x1a \x01(\x0b\x32\x16.Vault.ReusedPasswords\x12\x1a\n\x12removedUserFolders\x18\x1b \x03(\x0c\x12\x1c\n\x14removedSharedFolders\x18\x1c \x03(\x0c\x12\x45\n\x1eremovedUserFolderSharedFolders\x18\x1d \x03(\x0b\x32\x1d.Vault.UserFolderSharedFolder\x12=\n\x1aremovedSharedFolderFolders\x18\x1e \x03(\x0b\x32\x19.Vault.SharedFolderFolder\x12\x16\n\x0eremovedRecords\x18\x1f \x03(\x0c\x12-\n\x12removedRecordLinks\x18  \x03(\x0b\x32\x11.Vault.RecordLink\x12\x39\n\x18removedUserFolderRecords\x18! \x03(\x0b\x32\x17.Vault.UserFolderRecord\x12=\n\x1aremovedSharedFolderRecords\x18\" \x03(\x0b\x32\x19.Vault.SharedFolderRecord\x12I\n removedSharedFolderFolderRecords\x18# \x03(\x0b\x32\x1f.Vault.SharedFolderFolderRecord\x12\x39\n\x18removedSharedFolderUsers\x18$ \x03(\x0b\x32\x17.Vault.SharedFolderUser\x12\x39\n\x18removedSharedFolderTeams\x18% \x03(\x0b\x32\x17.Vault.SharedFolderTeam\x12\x14\n\x0cremovedTeams\x18& \x03(\x0c\x12&\n\x0cksmAppShares\x18\' \x03(\x0b\x32\x10.Vault.KsmChange\x12\'\n\rksmAppClients\x18( \x03(\x0b\x32\x10.Vault.KsmChange\x12\x30\n\x10shareInvitations\x18) \x03(\x0b\x32\x16.Vault.ShareInvitation\x12+\n\x0b\x64iagnostics\x18* \x01(\x0b\x32\x16.Vault.SyncDiagnostics\x12.\n\x0frecordRotations\x18+ \x03(\x0b\x32\x15.Vault.RecordRotation\x12\x1a\n\x05users\x18, \x03(\x0b\x32\x0b.Vault.User\x12\x14\n\x0cremovedUsers\x18- \x03(\x0c\"\x92\x01\n\nUserFolder\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12\x11\n\tparentUid\x18\x02 \x01(\x0c\x12\x15\n\ruserFolderKey\x18\x03 \x01(\x0c\x12\'\n\x07keyType\x18\x04 \x01(\x0e\x32\x16.Records.RecordKeyType\x12\x10\n\x08revision\x18\x05 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x06 \x01(\x0c\"\xc7\x02\n\x0cSharedFolder\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x10\n\x08revision\x18\x02 \x01(\x03\x12\x17\n\x0fsharedFolderKey\x18\x03 \x01(\x0c\x12\'\n\x07keyType\x18\x04 \x01(\x0e\x32\x16.Records.RecordKeyType\x12\x0c\n\x04\x64\x61ta\x18\x05 \x01(\x0c\x12\x1c\n\x14\x64\x65\x66\x61ultManageRecords\x18\x06 \x01(\x08\x12\x1a\n\x12\x64\x65\x66\x61ultManageUsers\x18\x07 \x01(\x08\x12\x16\n\x0e\x64\x65\x66\x61ultCanEdit\x18\x08 \x01(\x08\x12\x19\n\x11\x64\x65\x66\x61ultCanReshare\x18\t \x01(\x08\x12\'\n\x0b\x63\x61\x63heStatus\x18\n \x01(\x0e\x32\x12.Vault.CacheStatus\x12\r\n\x05owner\x18\x0b \x01(\t\x12\x17\n\x0fownerAccountUid\x18\x0c \x01(\x0c\"V\n\x16UserFolderSharedFolder\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12\x17\n\x0fsharedFolderUid\x18\x02 \x01(\x0c\x12\x10\n\x08revision\x18\x03 \x01(\x03\"\xbb\x01\n\x12SharedFolderFolder\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x11\n\tfolderUid\x18\x02 \x01(\x0c\x12\x11\n\tparentUid\x18\x03 \x01(\x0c\x12\x1d\n\x15sharedFolderFolderKey\x18\x04 \x01(\x0c\x12\'\n\x07keyType\x18\x05 \x01(\x0e\x32\x16.Records.RecordKeyType\x12\x10\n\x08revision\x18\x06 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x07 \x01(\x0c\"l\n\x0fSharedFolderKey\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x17\n\x0fsharedFolderKey\x18\x02 \x01(\x0c\x12\'\n\x07keyType\x18\x03 \x01(\x0e\x32\x16.Records.RecordKeyType\"\x8e\x02\n\x04Team\x12\x0f\n\x07teamUid\x18\x01 \x01(\x0c\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07teamKey\x18\x03 \x01(\x0c\x12+\n\x0bteamKeyType\x18\x04 \x01(\x0e\x32\x16.Records.RecordKeyType\x12\x16\n\x0eteamPrivateKey\x18\x05 \x01(\x0c\x12\x14\n\x0crestrictEdit\x18\x06 \x01(\x08\x12\x15\n\rrestrictShare\x18\x07 \x01(\x08\x12\x14\n\x0crestrictView\x18\x08 \x01(\x08\x12\x1c\n\x14removedSharedFolders\x18\t \x03(\x0c\x12\x30\n\x10sharedFolderKeys\x18\n \x03(\x0b\x32\x16.Vault.SharedFolderKey\"\xbf\x01\n\x06Record\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x10\n\x08revision\x18\x02 \x01(\x03\x12\x0f\n\x07version\x18\x03 \x01(\x05\x12\x0e\n\x06shared\x18\x04 \x01(\x08\x12\x1a\n\x12\x63lientModifiedTime\x18\x05 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x06 \x01(\x0c\x12\r\n\x05\x65xtra\x18\x07 \x01(\x0c\x12\r\n\x05udata\x18\x08 \x01(\t\x12\x10\n\x08\x66ileSize\x18\t \x01(\x03\x12\x15\n\rthumbnailSize\x18\n \x01(\x03\"b\n\nRecordLink\x12\x17\n\x0fparentRecordUid\x18\x01 \x01(\x0c\x12\x16\n\x0e\x63hildRecordUid\x18\x02 \x01(\x0c\x12\x11\n\trecordKey\x18\x03 \x01(\x0c\x12\x10\n\x08revision\x18\x04 \x01(\x03\"J\n\x10UserFolderRecord\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\x12\x10\n\x08revision\x18\x03 \x01(\x03\"Y\n\x18SharedFolderFolderRecord\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x11\n\tfolderUid\x18\x02 \x01(\x0c\x12\x11\n\trecordUid\x18\x03 \x01(\x0c\"0\n\rNonSharedData\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\"\xc4\x01\n\x0eRecordMetaData\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\r\n\x05owner\x18\x02 \x01(\x08\x12\x11\n\trecordKey\x18\x03 \x01(\x0c\x12-\n\rrecordKeyType\x18\x04 \x01(\x0e\x32\x16.Records.RecordKeyType\x12\x10\n\x08\x63\x61nShare\x18\x05 \x01(\x08\x12\x0f\n\x07\x63\x61nEdit\x18\x06 \x01(\x08\x12\x17\n\x0fownerAccountUid\x18\x07 \x01(\x0c\x12\x12\n\nexpiration\x18\x08 \x01(\x03\"2\n\rSharingChange\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x0e\n\x06shared\x18\x02 \x01(\x08\">\n\x07Profile\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\x13\n\x0bprofileName\x18\x02 \x01(\t\x12\x10\n\x08revision\x18\x03 \x01(\x03\"+\n\nProfilePic\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x10\n\x08revision\x18\x02 \x01(\x03\"V\n\x11PendingTeamMember\x12\x18\n\x10\x65nterpriseUserId\x18\x01 \x01(\x03\x12\x15\n\ruserPublicKey\x18\x02 \x01(\x0c\x12\x10\n\x08teamUids\x18\x03 \x03(\x0c\"\x89\x01\n\x11\x42reachWatchRecord\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12.\n\x04type\x18\x03 \x01(\x0e\x32 .BreachWatch.BreachWatchInfoType\x12\x11\n\tscannedBy\x18\x04 \x01(\t\x12\x10\n\x08revision\x18\x05 \x01(\x03\"\xb4\x01\n\x08UserAuth\x12\x0b\n\x03uid\x18\x01 \x01(\x0c\x12,\n\tloginType\x18\x02 \x01(\x0e\x32\x19.Authentication.LoginType\x12\x0f\n\x07\x64\x65leted\x18\x03 \x01(\x08\x12\x12\n\niterations\x18\x04 \x01(\x05\x12\x0c\n\x04salt\x18\x05 \x01(\x0c\x12\x1a\n\x12\x65ncryptedClientKey\x18\x06 \x01(\x0c\x12\x10\n\x08revision\x18\x07 \x01(\x03\x12\x0c\n\x04name\x18\x08 \x01(\t\">\n\x17\x42reachWatchSecurityData\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x10\n\x08revision\x18\x02 \x01(\x03\"2\n\x0fReusedPasswords\x12\r\n\x05\x63ount\x18\x01 \x01(\x05\x12\x10\n\x08revision\x18\x02 \x01(\x03\"\xa3\x01\n\x12SharedFolderRecord\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\x12\x11\n\trecordKey\x18\x03 \x01(\x0c\x12\x10\n\x08\x63\x61nShare\x18\x04 \x01(\x08\x12\x0f\n\x07\x63\x61nEdit\x18\x05 \x01(\x08\x12\x17\n\x0fownerAccountUid\x18\x06 \x01(\x0c\x12\x12\n\nexpiration\x18\x07 \x01(\x03\"\x91\x01\n\x10SharedFolderUser\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x10\n\x08username\x18\x02 \x01(\t\x12\x15\n\rmanageRecords\x18\x03 \x01(\x08\x12\x13\n\x0bmanageUsers\x18\x04 \x01(\x08\x12\x12\n\naccountUid\x18\x05 \x01(\x0c\x12\x12\n\nexpiration\x18\x06 \x01(\x03\"\x8a\x01\n\x10SharedFolderTeam\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x0f\n\x07teamUid\x18\x02 \x01(\x0c\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x15\n\rmanageRecords\x18\x04 \x01(\x08\x12\x13\n\x0bmanageUsers\x18\x05 \x01(\x08\x12\x12\n\nexpiration\x18\x06 \x01(\x03\"D\n\tKsmChange\x12\x14\n\x0c\x61ppRecordUid\x18\x01 \x01(\x0c\x12\x10\n\x08\x64\x65tailId\x18\x02 \x01(\x0c\x12\x0f\n\x07removed\x18\x03 \x01(\x08\"#\n\x0fShareInvitation\x12\x10\n\x08username\x18\x01 \x01(\t\",\n\x04User\x12\x12\n\naccountUid\x18\x01 \x01(\x0c\x12\x10\n\x08username\x18\x02 \x01(\t\"{\n\x0fSyncDiagnostics\x12\x19\n\x11\x63ontinuationToken\x18\x01 \x01(\x0c\x12\x0e\n\x06userId\x18\x02 \x01(\x05\x12\x18\n\x10\x65nterpriseUserId\x18\x03 \x01(\x03\x12\x10\n\x08syncedTo\x18\x04 \x01(\x03\x12\x11\n\tsyncingTo\x18\x05 \x01(\x03\"\x9d\x01\n\x0eRecordRotation\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x10\n\x08revision\x18\x02 \x01(\x03\x12\x18\n\x10\x63onfigurationUid\x18\x03 \x01(\x0c\x12\x10\n\x08schedule\x18\x04 \x01(\t\x12\x15\n\rpwdComplexity\x18\x05 \x01(\x0c\x12\x10\n\x08\x64isabled\x18\x06 \x01(\x08\x12\x11\n\tparentUid\x18\x07 \x01(\x0c\"3\n\x1d\x42reachWatchGetSyncDataRequest\x12\x12\n\nrecordUids\x18\x01 \x03(\x0c\"\x97\x01\n\x1e\x42reachWatchGetSyncDataResponse\x12\x34\n\x12\x62reachWatchRecords\x18\x01 \x03(\x0b\x32\x18.Vault.BreachWatchRecord\x12?\n\x17\x62reachWatchSecurityData\x18\x02 \x03(\x0b\x32\x1e.Vault.BreachWatchSecurityData*\"\n\x0b\x43\x61\x63heStatus\x12\x08\n\x04KEEP\x10\x00\x12\t\n\x05\x43LEAR\x10\x01\x42!\n\x18\x63om.keepersecurity.protoB\x05Vaultb\x06proto3')

_CACHESTATUS = DESCRIPTOR.enum_types_by_name['CacheStatus']
CacheStatus = enum_type_wrapper.EnumTypeWrapper(_CACHESTATUS)
KEEP = 0
CLEAR = 1


_SYNCDOWNREQUEST = DESCRIPTOR.message_types_by_name['SyncDownRequest']
_SYNCDOWNRESPONSE = DESCRIPTOR.message_types_by_name['SyncDownResponse']
_USERFOLDER = DESCRIPTOR.message_types_by_name['UserFolder']
_SHAREDFOLDER = DESCRIPTOR.message_types_by_name['SharedFolder']
_USERFOLDERSHAREDFOLDER = DESCRIPTOR.message_types_by_name['UserFolderSharedFolder']
_SHAREDFOLDERFOLDER = DESCRIPTOR.message_types_by_name['SharedFolderFolder']
_SHAREDFOLDERKEY = DESCRIPTOR.message_types_by_name['SharedFolderKey']
_TEAM = DESCRIPTOR.message_types_by_name['Team']
_RECORD = DESCRIPTOR.message_types_by_name['Record']
_RECORDLINK = DESCRIPTOR.message_types_by_name['RecordLink']
_USERFOLDERRECORD = DESCRIPTOR.message_types_by_name['UserFolderRecord']
_SHAREDFOLDERFOLDERRECORD = DESCRIPTOR.message_types_by_name['SharedFolderFolderRecord']
_NONSHAREDDATA = DESCRIPTOR.message_types_by_name['NonSharedData']
_RECORDMETADATA = DESCRIPTOR.message_types_by_name['RecordMetaData']
_SHARINGCHANGE = DESCRIPTOR.message_types_by_name['SharingChange']
_PROFILE = DESCRIPTOR.message_types_by_name['Profile']
_PROFILEPIC = DESCRIPTOR.message_types_by_name['ProfilePic']
_PENDINGTEAMMEMBER = DESCRIPTOR.message_types_by_name['PendingTeamMember']
_BREACHWATCHRECORD = DESCRIPTOR.message_types_by_name['BreachWatchRecord']
_USERAUTH = DESCRIPTOR.message_types_by_name['UserAuth']
_BREACHWATCHSECURITYDATA = DESCRIPTOR.message_types_by_name['BreachWatchSecurityData']
_REUSEDPASSWORDS = DESCRIPTOR.message_types_by_name['ReusedPasswords']
_SHAREDFOLDERRECORD = DESCRIPTOR.message_types_by_name['SharedFolderRecord']
_SHAREDFOLDERUSER = DESCRIPTOR.message_types_by_name['SharedFolderUser']
_SHAREDFOLDERTEAM = DESCRIPTOR.message_types_by_name['SharedFolderTeam']
_KSMCHANGE = DESCRIPTOR.message_types_by_name['KsmChange']
_SHAREINVITATION = DESCRIPTOR.message_types_by_name['ShareInvitation']
_USER = DESCRIPTOR.message_types_by_name['User']
_SYNCDIAGNOSTICS = DESCRIPTOR.message_types_by_name['SyncDiagnostics']
_RECORDROTATION = DESCRIPTOR.message_types_by_name['RecordRotation']
_BREACHWATCHGETSYNCDATAREQUEST = DESCRIPTOR.message_types_by_name['BreachWatchGetSyncDataRequest']
_BREACHWATCHGETSYNCDATARESPONSE = DESCRIPTOR.message_types_by_name['BreachWatchGetSyncDataResponse']
SyncDownRequest = _reflection.GeneratedProtocolMessageType('SyncDownRequest', (_message.Message,), {
  'DESCRIPTOR' : _SYNCDOWNREQUEST,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.SyncDownRequest)
  })
_sym_db.RegisterMessage(SyncDownRequest)

SyncDownResponse = _reflection.GeneratedProtocolMessageType('SyncDownResponse', (_message.Message,), {
  'DESCRIPTOR' : _SYNCDOWNRESPONSE,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.SyncDownResponse)
  })
_sym_db.RegisterMessage(SyncDownResponse)

UserFolder = _reflection.GeneratedProtocolMessageType('UserFolder', (_message.Message,), {
  'DESCRIPTOR' : _USERFOLDER,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.UserFolder)
  })
_sym_db.RegisterMessage(UserFolder)

SharedFolder = _reflection.GeneratedProtocolMessageType('SharedFolder', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDER,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.SharedFolder)
  })
_sym_db.RegisterMessage(SharedFolder)

UserFolderSharedFolder = _reflection.GeneratedProtocolMessageType('UserFolderSharedFolder', (_message.Message,), {
  'DESCRIPTOR' : _USERFOLDERSHAREDFOLDER,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.UserFolderSharedFolder)
  })
_sym_db.RegisterMessage(UserFolderSharedFolder)

SharedFolderFolder = _reflection.GeneratedProtocolMessageType('SharedFolderFolder', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERFOLDER,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.SharedFolderFolder)
  })
_sym_db.RegisterMessage(SharedFolderFolder)

SharedFolderKey = _reflection.GeneratedProtocolMessageType('SharedFolderKey', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERKEY,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.SharedFolderKey)
  })
_sym_db.RegisterMessage(SharedFolderKey)

Team = _reflection.GeneratedProtocolMessageType('Team', (_message.Message,), {
  'DESCRIPTOR' : _TEAM,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.Team)
  })
_sym_db.RegisterMessage(Team)

Record = _reflection.GeneratedProtocolMessageType('Record', (_message.Message,), {
  'DESCRIPTOR' : _RECORD,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.Record)
  })
_sym_db.RegisterMessage(Record)

RecordLink = _reflection.GeneratedProtocolMessageType('RecordLink', (_message.Message,), {
  'DESCRIPTOR' : _RECORDLINK,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.RecordLink)
  })
_sym_db.RegisterMessage(RecordLink)

UserFolderRecord = _reflection.GeneratedProtocolMessageType('UserFolderRecord', (_message.Message,), {
  'DESCRIPTOR' : _USERFOLDERRECORD,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.UserFolderRecord)
  })
_sym_db.RegisterMessage(UserFolderRecord)

SharedFolderFolderRecord = _reflection.GeneratedProtocolMessageType('SharedFolderFolderRecord', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERFOLDERRECORD,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.SharedFolderFolderRecord)
  })
_sym_db.RegisterMessage(SharedFolderFolderRecord)

NonSharedData = _reflection.GeneratedProtocolMessageType('NonSharedData', (_message.Message,), {
  'DESCRIPTOR' : _NONSHAREDDATA,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.NonSharedData)
  })
_sym_db.RegisterMessage(NonSharedData)

RecordMetaData = _reflection.GeneratedProtocolMessageType('RecordMetaData', (_message.Message,), {
  'DESCRIPTOR' : _RECORDMETADATA,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.RecordMetaData)
  })
_sym_db.RegisterMessage(RecordMetaData)

SharingChange = _reflection.GeneratedProtocolMessageType('SharingChange', (_message.Message,), {
  'DESCRIPTOR' : _SHARINGCHANGE,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.SharingChange)
  })
_sym_db.RegisterMessage(SharingChange)

Profile = _reflection.GeneratedProtocolMessageType('Profile', (_message.Message,), {
  'DESCRIPTOR' : _PROFILE,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.Profile)
  })
_sym_db.RegisterMessage(Profile)

ProfilePic = _reflection.GeneratedProtocolMessageType('ProfilePic', (_message.Message,), {
  'DESCRIPTOR' : _PROFILEPIC,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.ProfilePic)
  })
_sym_db.RegisterMessage(ProfilePic)

PendingTeamMember = _reflection.GeneratedProtocolMessageType('PendingTeamMember', (_message.Message,), {
  'DESCRIPTOR' : _PENDINGTEAMMEMBER,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.PendingTeamMember)
  })
_sym_db.RegisterMessage(PendingTeamMember)

BreachWatchRecord = _reflection.GeneratedProtocolMessageType('BreachWatchRecord', (_message.Message,), {
  'DESCRIPTOR' : _BREACHWATCHRECORD,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.BreachWatchRecord)
  })
_sym_db.RegisterMessage(BreachWatchRecord)

UserAuth = _reflection.GeneratedProtocolMessageType('UserAuth', (_message.Message,), {
  'DESCRIPTOR' : _USERAUTH,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.UserAuth)
  })
_sym_db.RegisterMessage(UserAuth)

BreachWatchSecurityData = _reflection.GeneratedProtocolMessageType('BreachWatchSecurityData', (_message.Message,), {
  'DESCRIPTOR' : _BREACHWATCHSECURITYDATA,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.BreachWatchSecurityData)
  })
_sym_db.RegisterMessage(BreachWatchSecurityData)

ReusedPasswords = _reflection.GeneratedProtocolMessageType('ReusedPasswords', (_message.Message,), {
  'DESCRIPTOR' : _REUSEDPASSWORDS,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.ReusedPasswords)
  })
_sym_db.RegisterMessage(ReusedPasswords)

SharedFolderRecord = _reflection.GeneratedProtocolMessageType('SharedFolderRecord', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERRECORD,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.SharedFolderRecord)
  })
_sym_db.RegisterMessage(SharedFolderRecord)

SharedFolderUser = _reflection.GeneratedProtocolMessageType('SharedFolderUser', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERUSER,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.SharedFolderUser)
  })
_sym_db.RegisterMessage(SharedFolderUser)

SharedFolderTeam = _reflection.GeneratedProtocolMessageType('SharedFolderTeam', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERTEAM,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.SharedFolderTeam)
  })
_sym_db.RegisterMessage(SharedFolderTeam)

KsmChange = _reflection.GeneratedProtocolMessageType('KsmChange', (_message.Message,), {
  'DESCRIPTOR' : _KSMCHANGE,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.KsmChange)
  })
_sym_db.RegisterMessage(KsmChange)

ShareInvitation = _reflection.GeneratedProtocolMessageType('ShareInvitation', (_message.Message,), {
  'DESCRIPTOR' : _SHAREINVITATION,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.ShareInvitation)
  })
_sym_db.RegisterMessage(ShareInvitation)

User = _reflection.GeneratedProtocolMessageType('User', (_message.Message,), {
  'DESCRIPTOR' : _USER,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.User)
  })
_sym_db.RegisterMessage(User)

SyncDiagnostics = _reflection.GeneratedProtocolMessageType('SyncDiagnostics', (_message.Message,), {
  'DESCRIPTOR' : _SYNCDIAGNOSTICS,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.SyncDiagnostics)
  })
_sym_db.RegisterMessage(SyncDiagnostics)

RecordRotation = _reflection.GeneratedProtocolMessageType('RecordRotation', (_message.Message,), {
  'DESCRIPTOR' : _RECORDROTATION,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.RecordRotation)
  })
_sym_db.RegisterMessage(RecordRotation)

BreachWatchGetSyncDataRequest = _reflection.GeneratedProtocolMessageType('BreachWatchGetSyncDataRequest', (_message.Message,), {
  'DESCRIPTOR' : _BREACHWATCHGETSYNCDATAREQUEST,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.BreachWatchGetSyncDataRequest)
  })
_sym_db.RegisterMessage(BreachWatchGetSyncDataRequest)

BreachWatchGetSyncDataResponse = _reflection.GeneratedProtocolMessageType('BreachWatchGetSyncDataResponse', (_message.Message,), {
  'DESCRIPTOR' : _BREACHWATCHGETSYNCDATARESPONSE,
  '__module__' : 'SyncDown_pb2'
  # @@protoc_insertion_point(class_scope:Vault.BreachWatchGetSyncDataResponse)
  })
_sym_db.RegisterMessage(BreachWatchGetSyncDataResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030com.keepersecurity.protoB\005Vault'
  _CACHESTATUS._serialized_start=5820
  _CACHESTATUS._serialized_end=5854
  _SYNCDOWNREQUEST._serialized_start=76
  _SYNCDOWNREQUEST._serialized_end=120
  _SYNCDOWNRESPONSE._serialized_start=123
  _SYNCDOWNRESPONSE._serialized_end=2180
  _USERFOLDER._serialized_start=2183
  _USERFOLDER._serialized_end=2329
  _SHAREDFOLDER._serialized_start=2332
  _SHAREDFOLDER._serialized_end=2659
  _USERFOLDERSHAREDFOLDER._serialized_start=2661
  _USERFOLDERSHAREDFOLDER._serialized_end=2747
  _SHAREDFOLDERFOLDER._serialized_start=2750
  _SHAREDFOLDERFOLDER._serialized_end=2937
  _SHAREDFOLDERKEY._serialized_start=2939
  _SHAREDFOLDERKEY._serialized_end=3047
  _TEAM._serialized_start=3050
  _TEAM._serialized_end=3320
  _RECORD._serialized_start=3323
  _RECORD._serialized_end=3514
  _RECORDLINK._serialized_start=3516
  _RECORDLINK._serialized_end=3614
  _USERFOLDERRECORD._serialized_start=3616
  _USERFOLDERRECORD._serialized_end=3690
  _SHAREDFOLDERFOLDERRECORD._serialized_start=3692
  _SHAREDFOLDERFOLDERRECORD._serialized_end=3781
  _NONSHAREDDATA._serialized_start=3783
  _NONSHAREDDATA._serialized_end=3831
  _RECORDMETADATA._serialized_start=3834
  _RECORDMETADATA._serialized_end=4030
  _SHARINGCHANGE._serialized_start=4032
  _SHARINGCHANGE._serialized_end=4082
  _PROFILE._serialized_start=4084
  _PROFILE._serialized_end=4146
  _PROFILEPIC._serialized_start=4148
  _PROFILEPIC._serialized_end=4191
  _PENDINGTEAMMEMBER._serialized_start=4193
  _PENDINGTEAMMEMBER._serialized_end=4279
  _BREACHWATCHRECORD._serialized_start=4282
  _BREACHWATCHRECORD._serialized_end=4419
  _USERAUTH._serialized_start=4422
  _USERAUTH._serialized_end=4602
  _BREACHWATCHSECURITYDATA._serialized_start=4604
  _BREACHWATCHSECURITYDATA._serialized_end=4666
  _REUSEDPASSWORDS._serialized_start=4668
  _REUSEDPASSWORDS._serialized_end=4718
  _SHAREDFOLDERRECORD._serialized_start=4721
  _SHAREDFOLDERRECORD._serialized_end=4884
  _SHAREDFOLDERUSER._serialized_start=4887
  _SHAREDFOLDERUSER._serialized_end=5032
  _SHAREDFOLDERTEAM._serialized_start=5035
  _SHAREDFOLDERTEAM._serialized_end=5173
  _KSMCHANGE._serialized_start=5175
  _KSMCHANGE._serialized_end=5243
  _SHAREINVITATION._serialized_start=5245
  _SHAREINVITATION._serialized_end=5280
  _USER._serialized_start=5282
  _USER._serialized_end=5326
  _SYNCDIAGNOSTICS._serialized_start=5328
  _SYNCDIAGNOSTICS._serialized_end=5451
  _RECORDROTATION._serialized_start=5454
  _RECORDROTATION._serialized_end=5611
  _BREACHWATCHGETSYNCDATAREQUEST._serialized_start=5613
  _BREACHWATCHGETSYNCDATAREQUEST._serialized_end=5664
  _BREACHWATCHGETSYNCDATARESPONSE._serialized_start=5667
  _BREACHWATCHGETSYNCDATARESPONSE._serialized_end=5818
# @@protoc_insertion_point(module_scope)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import getpass
import io
import logging
import os
import re
import uuid
from typing import Dict
from xml.sax.saxutils import escape

from pykeepass import PyKeePass
from pykeepass.exceptions import CredentialsError
from pykeepass.attachment import Attachment as KeepassAttachment
from pykeepass.group import Group

from ..importer import path_components, PathDelimiter, BaseFileImporter, BaseExporter, \
    Record, Folder, SharedFolder, BytesAttachment, RecordField
from ... import utils

_REFERENCE = r'\{REF:([TUPAN])@([IT]):([^\}]+)\}'


class XmlUtils(object):
    @staticmethod
    def escape_string(plain):   # type: (str) -> str
        if not plain:
            return ''
        output = escape(plain)
        return output.replace('\'', '&apos;').replace('\"', '&quot;')


class KeepassImporter(BaseFileImporter):

    @staticmethod
    def get_folder(group):  # type: (Group) -> str
        g = group
        path = ''
        comp = ''
        while isinstance(g, Group):
            if comp:
                if len(path) > 0:
                    path = PathDelimiter + path
                path = comp + path
                comp = None

            nm = g.name
            if nm:
                comp = nm.replace(PathDelimiter, PathDelimiter*2)
            g = g.group
        return path

    def do_import(self, filename, **kwargs):
        password = getpass.getpass(prompt='...' + 'Keepass Password'.rjust(20) + ': ', stream=None)
        print('Press Enter if your Keepass file is not protected with a key file')
        keyfile = input('...' + 'Path to Key file'.rjust(20) + ': ')
        if keyfile:
            keyfile = os.path.expanduser(keyfile)
        else:
            keyfile = None
        try:
            with PyKeePass(filename, password=password, keyfile=keyfile) as kdb:
                root = kdb.root_group
                if not root:
                    return
                groups = [root]  # type: list
                pos = 0
                while pos < len(groups):
                    g = groups[pos]
                    groups.extend(g.subgroups)
                    pos = pos + 1

                records = []
                for group in groups:
                    entries = group.entries
                    if len(entries) > 0:
                        folder = KeepassImporter.get_folder(group)
                        for entry in entries:
                            record = Record()
                            fol = Folder()
                            fol.path = folder
                            record.folders = [fol]
                            if entry.uuid:
                                record.uid = utils.base64_url_encode(entry.uuid.bytes)
                            if entry.title:
                                record.title = entry.title
                            if entry.username:
                                record.login = entry.username
                            if entry.password:
                                record.password = entry.password
                            if entry.url:
                                record.login_url = entry.url
                            if entry.notes:
                                record.notes = entry.notes
                            for key, value in entry.custom_properties.items():
                                if key == '$type':
                                    record.type = value
                                    continue
                                rest, sep, no = key.rpartition('#')
                                if sep:
                                    if no.isdigit():
                                        key = rest
                                if key.startswith('$'):
                                    ftype, sep, flabel = key.partition(':')
                                    field_type = ftype[1:]
                                    field_label = flabel
                                else:
                                    field_type = ''
                                    field_label = key
                                field = RecordField()
                                field.type = field_type
                                field.label = field_label
                                field.value = KeepassImporter.import_field(field_type, value)
                                record.fields.append(field)

                            if entry.attachments:
                                for a in entry.attachments:
                                    if isinstance(a, KeepassAttachment):
                                        if record.attachments is None:
                                            record.attachments = []
                                        record.attachments.append(BytesAttachment(a.filename, a.binary))

                            records.append(record)
                id_map = None
                title_map = None
                ref_re = re.compile(_REFERENCE, re.IGNORECASE)

                def resolve_references(text):
                    nonlocal id_map
                    nonlocal title_map

                    if not text:
                        return text
                    matches = list(ref_re.finditer(text))
                    if not matches:
                        return text
                    values = []
                    for match in matches:
                        comps = match.groups()
                        val = match.group(0)
                        if len(comps) == 3:
                            rec = None
                            if comps[1].upper() == 'I':
                                if id_map is None:
                                    id_map = {}
                                    for r in records:
                                        if r.uid:
                                            id_map[r.uid] = r
                                uid = uuid.UUID(comps[2])
                                if uid:
                                    ref_id = utils.base64_url_encode(uid.bytes)
                                    rec = id_map.get(ref_id)
                            elif comps[1].upper() == 'T':
                                if title_map is None:
                                    title_map = {}
                                    for r in records:
                                        if r.title:
                                            title_map[r.title] = r
                                rec = title_map.get(comps[2])
                            if rec:
                                field_id = comps[0].upper()
                                if field_id == 'T':
                                    val = rec.title
                                elif field_id == 'U':
                                    val = rec.login
                                elif field_id == 'P':
                                    val = rec.password
                                elif field_id == 'A':
                                    val = rec.login_url
                                elif field_id == 'N':
                                    val = rec.notes
                        values.append(val or '')

                        idx = list(range(len(matches)))
                        idx.reverse()
                        for index in idx:
                            match = matches[index]
                            val = values[index] if index < len(values) else ''
                            text = text[:match.start()] + val + text[match.end():]
                        return text

                for record in records:
                    try:
                        record.login = resolve_references(record.login)
                        record.password = resolve_references(record.password)
                        record.login_url = resolve_references(record.login_url)
                        record.notes = resolve_references(record.notes)
                        for field in record.fields:
                            field.value = resolve_references(field.value)
                    except Exception as e:
                        logging.debug(e)
                    yield record
        except CredentialsError:
            logging.warning('Invalid Keepass credentials')

    def extension(self):
        return 'kdbx'


class KeepassExporter(BaseExporter, XmlUtils):
    @staticmethod
    def to_keepass_value(keeper_value):  # type: (any) -> str
        if not keeper_value:
            return ''
        if isinstance(keeper_value, list):
            return ','.join((KeepassExporter.to_keepass_value(x) for x in keeper_value))
        elif isinstance(keeper_value, dict):
            return ';\n'.join((f'{k}:{KeepassExporter.to_keepass_value(v)}' for k, v in keeper_value.items()))
        else:
            return str(keeper_value)

    def do_export(self, filename, records, file_password=None):
        master_password = file_password
        if not master_password:
            print('Choose password for your Keepass file')
            master_password = getpass.getpass(prompt='...' + 'Keepass Password'.rjust(20) + ': ', stream=None)

        sfs = []  # type: list[SharedFolder]
        rs = []   # type: list[Record]
        for x in records:
            if isinstance(x, Record):
                rs.append(x)
            elif isinstance(x, SharedFolder):
                sfs.append(x)

        template_file = os.path.join(os.path.dirname(__file__), 'template.kdbx')

        with PyKeePass(template_file, password='111111') as kdb:
            kdb.password = master_password
            root = kdb.root_group

            for r in rs:
                try:
                    node = root
                    if r.folders:
                        fol = r.folders[0]
                        for is_shared in [True, False]:
                            path = fol.domain if is_shared else fol.path
                            if path:
                                comps = list(path_components(path))
                                for i in range(len(comps)):
                                    comp = comps[i]
                                    sub_node = next((x for x in node.subgroups if x.name == comp), None)
                                    if sub_node is None:
                                        sub_node = kdb.add_group(node, comp)
                                    node = sub_node
                    entry = None
                    entries = node.entries
                    for en in entries:
                        if en.title == r.title and en.username == r.login and en.password == r.password:
                            entry = en
                            break

                    if entry is None:
                        entry = kdb.add_entry(node, title=r.title or '', username=r.login or '',
                                              password=r.password or '', url=r.login_url or '',
                                              notes=r.notes or '')
                        if r.uid:
                            entry.UUID = uuid.UUID(bytes=utils.base64_url_decode(r.uid))
                    if r.type:
                        entry.set_custom_property('$type', r.type)
                    if r.fields:
                        custom_names = {}   # type: Dict[str, int]
                        for cf in r.fields:
                            if cf.type and cf.label:
                                title = f'${cf.type}:{cf.label}'
                            elif cf.type:
                                title = f'${cf.type}'
                            else:
                                title = cf.label or ''
                            if title in custom_names:
                                no = custom_names[title]
                                no += 1
                                custom_names[title] = no
                                title = f'{title}#{no}'
                            else:
                                custom_names[title] = 1
                            entry.set_custom_property(title, self.to_keepass_value(BaseExporter.export_field(cf.type, cf.value)))

                    if r.attachments:
                        for atta in r.attachments:
                            if atta.size < self.max_size:
                                with atta.open() as ina, io.BytesIO() as outa:
                                    buffer = bytearray(10240)
                                    view = memoryview(buffer)
                                    while True:
                                        bytes_read = ina.readinto(view)
                                        if bytes_read == 0:
                                            break
                                        outa.write(view[0:bytes_read])
                                    del view
                                    del buffer
                                    outa.flush()
                                    binary = outa.getvalue()

                                if binary:
                                    binary_id = kdb.add_binary(binary, compressed=True, protected=False)
                                    entry.add_attachment(binary_id, atta.name)
                            else:
                                scale = ''
                                msize = self.max_size
                                if msize > 1024 ** 3:
                                    scale = 'G'
                                    msize //= 1024 ** 3
                                elif msize > 1024 ** 2:
                                    scale = 'M'
                                    msize //= 1024 ** 2
                                elif msize > 1024:
                                    scale = 'K'
                                    msize //= 1024
                                logging.warning(
                                    'File \'{0}\' was skipped because it exceeds the {1}{2} file size limit.'.format(atta.name, msize, scale))
                except Exception as e:
                    logging.debug(e)

            kdb.save(filename=filename)

    def has_shared_folders(self):
        return True

    def has_attachments(self):
        return True

    def extension(self):
        return 'kdbx'

    def supports_v3_record(self):
        return False



#  _  __  
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|            
#
# Keeper Commander 
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import logging
import os
import threading
import time
from typing import Optional, Callable, Union

from fido2.client import Fido2Client, WindowsClient, ClientError
from fido2.ctap import CtapError
from fido2.hid import CtapHidDevice
from fido2.webauthn import PublicKeyCredentialRequestOptions, AuthenticatorAssertionResponse
from prompt_toolkit import PromptSession

from .. import utils


def verify_rp_id_none(rp_id, origin):
    return True


def yubikey_authenticate(request):  # type: (dict) -> Optional[dict]
    auth_func = None    # type: Optional[Callable[[], Union[AuthenticatorAssertionResponse, dict, None]]]
    evt = threading.Event()
    response = None  # type: Optional[str]

    if 'publicKeyCredentialRequestOptions' in request:  # WebAuthN
        origin = ''
        options = request['publicKeyCredentialRequestOptions']
        if 'extensions' in options:
            extensions = options['extensions']
            origin = extensions.get('appid') or ''

        credentials = options.get('allowCredentials') or []
        for c in credentials:
            if isinstance(c.get('id'), str):
                c['id'] = utils.base64_url_decode(c['id'])

        rq_options = PublicKeyCredentialRequestOptions(utils.base64_url_decode(options['challenge']),
                                                       rp_id=options['rpId'], user_verification='discouraged',
                                                       allow_credentials=credentials)

        if WindowsClient.is_available():
            client = WindowsClient(origin, verify=verify_rp_id_none)
        else:
            dev = next(CtapHidDevice.list_devices(), None)
            if not dev:
                logging.warning("No Security Key detected")
                return
            client = Fido2Client(dev, origin, verify=verify_rp_id_none)

        def auth_func():
            nonlocal response
            nonlocal rq_options
            attempt = 0
            while attempt < 2:
                attempt += 1
                try:
                    rs = client.get_assertion(rq_options, event=evt)
                    response = rs.get_response(0)
                    break
                except ClientError as err:
                    if isinstance(err.cause, CtapError) and attempt == 1:
                        if err.cause.code == CtapError.ERR.NO_CREDENTIALS:
                            print('\n\nKeeper Security stopped supporting U2F security keys starting February 2022.\n'
                                  'If you registered your security key prior to this date please re-register it within the Web Vault.\n'
                                  'For information on using security keys with Keeper see the documentation: \n'
                                  'https://docs.keeper.io/enterprise-guide/two-factor-authentication#security-keys-fido-webauthn\n'
                                  'Commander will use the fallback security key authentication method.\n\n'
                                  'To use your Yubikey with Commander, please touch the flashing Security key one more time.\n')
                            rq_options = PublicKeyCredentialRequestOptions(
                                utils.base64_url_decode(options['challenge']), rp_id=origin, user_verification='discouraged',
                                allow_credentials=credentials)
                            continue
                    raise err
    else:
        logging.warning('Invalid Security Key request')
        return

    prompt_session = None

    def func():
        nonlocal prompt_session
        nonlocal evt
        try:
            time.sleep(0.1)
            auth_func()
        except:
            pass
        if prompt_session:
            evt = None
            prompt_session.app.exit()
        elif evt:
            print('\npress Enter to resume...')

    th = threading.Thread(target=func)
    th.start()
    try:
        prompt = 'Touch the flashing Security key to authenticate or press Enter to resume with the primary two factor authentication...'
        if os.isatty(0) and os.isatty(1):
            prompt_session = PromptSession(multiline=False, complete_while_typing=False)
            prompt_session.prompt(prompt)
            prompt_session = None
        else:
            input(prompt)
    except KeyboardInterrupt:
        prompt_session = None
    if evt:
        evt.set()
        evt = None
    th.join()

    return response



from typing import Optional, List

from .vault import TypedRecord

class TypedRecordFacade:
    record: Optional[TypedRecord]
    title: str
    notes: str
    def load_typed_fields(self) -> None: ...


class FileRefRecordFacade(TypedRecordFacade):
    @property
    def file_ref(self) -> List[str]: ...

class LoginRecordFacade(FileRefRecordFacade):
    login: str
    password: str
    url: str
    oneTimeCode: str
